# -*- coding: utf-8 -*-
"""dcmstack allows series of DICOM images to be stacked into multi-dimensional arrays."""

import os
from os import path as op
import string
import errno
from glob import glob

import nibabel as nb
import imghdr

from .base import (
    TraitedSpec,
    DynamicTraitedSpec,
    InputMultiPath,
    File,
    Directory,
    traits,
    BaseInterface,
    isdefined,
    Undefined,
)

have_dcmstack = True
try:
    import dicom
    import dcmstack
    from dcmstack.dcmmeta import NiftiWrapper
except ImportError:
    have_dcmstack = False


def sanitize_path_comp(path_comp):
    result = []
    for char in path_comp:
        if char not in string.letters + string.digits + "-_.":
            result.append("_")
        else:
            result.append(char)
    return "".join(result)


class NiftiGeneratorBaseInputSpec(TraitedSpec):
    out_format = traits.Str(
        desc="String which can be formatted with "
        "meta data to create the output filename(s)"
    )
    out_ext = traits.Str(".nii.gz", usedefault=True, desc="Determines output file type")
    out_path = Directory(desc="output path, current working directory if not set")


class NiftiGeneratorBase(BaseInterface):
    """Base class for interfaces that produce Nifti files, potentially with
    embedded meta data."""

    def _get_out_path(self, meta, idx=None):
        """Return the output path for the gernerated Nifti."""
        if self.inputs.out_format:
            out_fmt = self.inputs.out_format
        else:
            # If no out_format is specified, use a sane default that will work
            # with the provided meta data.
            out_fmt = []
            if idx is not None:
                out_fmt.append("%03d" % idx)
            if "SeriesNumber" in meta:
                out_fmt.append("%(SeriesNumber)03d")
            if "ProtocolName" in meta:
                out_fmt.append("%(ProtocolName)s")
            elif "SeriesDescription" in meta:
                out_fmt.append("%(SeriesDescription)s")
            else:
                out_fmt.append("sequence")
            out_fmt = "-".join(out_fmt)
        out_fn = (out_fmt % meta) + self.inputs.out_ext
        out_fn = sanitize_path_comp(out_fn)

        out_path = os.getcwd()
        if isdefined(self.inputs.out_path):
            out_path = op.abspath(self.inputs.out_path)

            # now, mkdir -p $out_path
            try:
                os.makedirs(out_path)
            except OSError as exc:  # Python >2.5
                if exc.errno == errno.EEXIST and op.isdir(out_path):
                    pass
                else:
                    raise

        return op.join(out_path, out_fn)


class DcmStackInputSpec(NiftiGeneratorBaseInputSpec):
    dicom_files = traits.Either(
        InputMultiPath(File(exists=True)),
        Directory(exists=True),
        traits.Str(),
        mandatory=True,
    )
    embed_meta = traits.Bool(desc="Embed DICOM meta data into result")
    exclude_regexes = traits.List(
        desc="Meta data to exclude, suplementing " "any default exclude filters"
    )
    include_regexes = traits.List(
        desc="Meta data to include, overriding any " "exclude filters"
    )
    force_read = traits.Bool(
        True, usedefault=True, desc=("Force reading files without DICM marker")
    )


class DcmStackOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class DcmStack(NiftiGeneratorBase):
    """Create one Nifti file from a set of DICOM files. Can optionally embed
    meta data.

    Example
    -------

    >>> from nipype.interfaces.dcmstack import DcmStack
    >>> stacker = DcmStack()
    >>> stacker.inputs.dicom_files = 'path/to/series/'
    >>> stacker.run() # doctest: +SKIP
    >>> result.outputs.out_file # doctest: +SKIP
    '/path/to/cwd/sequence.nii.gz'
    """

    input_spec = DcmStackInputSpec
    output_spec = DcmStackOutputSpec

    def _get_filelist(self, trait_input):
        if isinstance(trait_input, (str, bytes)):
            if op.isdir(trait_input):
                return glob(op.join(trait_input, "*.dcm"))
            else:
                return glob(trait_input)

        return trait_input

    def _run_interface(self, runtime):
        src_paths = self._get_filelist(self.inputs.dicom_files)
        include_regexes = dcmstack.default_key_incl_res
        if isdefined(self.inputs.include_regexes):
            include_regexes += self.inputs.include_regexes
        exclude_regexes = dcmstack.default_key_excl_res
        if isdefined(self.inputs.exclude_regexes):
            exclude_regexes += self.inputs.exclude_regexes
        meta_filter = dcmstack.make_key_regex_filter(exclude_regexes, include_regexes)
        stack = dcmstack.DicomStack(meta_filter=meta_filter)
        for src_path in src_paths:
            if not imghdr.what(src_path) == "gif":
                src_dcm = dicom.read_file(src_path, force=self.inputs.force_read)
                stack.add_dcm(src_dcm)
        nii = stack.to_nifti(embed_meta=True)
        nw = NiftiWrapper(nii)
        self.out_path = self._get_out_path(
            nw.meta_ext.get_class_dict(("global", "const"))
        )
        if not self.inputs.embed_meta:
            nw.remove_extension()
        nb.save(nii, self.out_path)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = self.out_path
        return outputs


class GroupAndStackOutputSpec(TraitedSpec):
    out_list = traits.List(desc="List of output nifti files")


class GroupAndStack(DcmStack):
    """Create (potentially) multiple Nifti files for a set of DICOM files.
    """

    input_spec = DcmStackInputSpec
    output_spec = GroupAndStackOutputSpec

    def _run_interface(self, runtime):
        src_paths = self._get_filelist(self.inputs.dicom_files)
        stacks = dcmstack.parse_and_stack(src_paths)

        self.out_list = []
        for key, stack in list(stacks.items()):
            nw = NiftiWrapper(stack.to_nifti(embed_meta=True))
            const_meta = nw.meta_ext.get_class_dict(("global", "const"))
            out_path = self._get_out_path(const_meta)
            if not self.inputs.embed_meta:
                nw.remove_extension()
            nb.save(nw.nii_img, out_path)
            self.out_list.append(out_path)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_list"] = self.out_list
        return outputs


class LookupMetaInputSpec(TraitedSpec):
    in_file = File(mandatory=True, exists=True, desc="The input Nifti file")
    meta_keys = traits.Either(
        traits.List(),
        traits.Dict(),
        mandatory=True,
        desc=(
            "List of meta data keys to lookup, or a "
            "dict where keys specify the meta data "
            "keys to lookup and the values specify "
            "the output names"
        ),
    )


class LookupMeta(BaseInterface):
    """Lookup meta data values from a Nifti with embedded meta data.

    Example
    -------

    >>> from nipype.interfaces import dcmstack
    >>> lookup = dcmstack.LookupMeta()
    >>> lookup.inputs.in_file = 'functional.nii'
    >>> lookup.inputs.meta_keys = {'RepetitionTime' : 'TR', \
                                   'EchoTime' : 'TE'}
    >>> result = lookup.run() # doctest: +SKIP
    >>> result.outputs.TR # doctest: +SKIP
    9500.0
    >>> result.outputs.TE # doctest: +SKIP
    95.0
    """

    input_spec = LookupMetaInputSpec
    output_spec = DynamicTraitedSpec

    def _make_name_map(self):
        if isinstance(self.inputs.meta_keys, list):
            self._meta_keys = {}
            for key in self.inputs.meta_keys:
                self._meta_keys[key] = key
        else:
            self._meta_keys = self.inputs.meta_keys

    def _outputs(self):
        self._make_name_map()
        outputs = super(LookupMeta, self)._outputs()
        undefined_traits = {}
        for out_name in list(self._meta_keys.values()):
            outputs.add_trait(out_name, traits.Any)
            undefined_traits[out_name] = Undefined
        outputs.trait_set(trait_change_notify=False, **undefined_traits)
        # Not sure why this is needed
        for out_name in list(self._meta_keys.values()):
            _ = getattr(outputs, out_name)
        return outputs

    def _run_interface(self, runtime):
        # If the 'meta_keys' input is a list, covert it to a dict
        self._make_name_map()
        nw = NiftiWrapper.from_filename(self.inputs.in_file)
        self.result = {}
        for meta_key, out_name in list(self._meta_keys.items()):
            self.result[out_name] = nw.meta_ext.get_values(meta_key)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs.update(self.result)
        return outputs


class CopyMetaInputSpec(TraitedSpec):
    src_file = File(mandatory=True, exists=True)
    dest_file = File(mandatory=True, exists=True)
    include_classes = traits.List(
        desc="List of specific meta data "
        "classifications to include. If not "
        "specified include everything."
    )
    exclude_classes = traits.List(
        desc="List of meta data " "classifications to exclude"
    )


class CopyMetaOutputSpec(TraitedSpec):
    dest_file = File(exists=True)


class CopyMeta(BaseInterface):
    """Copy meta data from one Nifti file to another. Useful for preserving
    meta data after some processing steps."""

    input_spec = CopyMetaInputSpec
    output_spec = CopyMetaOutputSpec

    def _run_interface(self, runtime):
        src_nii = nb.load(self.inputs.src_file)
        src = NiftiWrapper(src_nii, make_empty=True)
        dest_nii = nb.load(self.inputs.dest_file)
        dest = NiftiWrapper(dest_nii, make_empty=True)
        classes = src.meta_ext.get_valid_classes()
        if self.inputs.include_classes:
            classes = [cls for cls in classes if cls in self.inputs.include_classes]
        if self.inputs.exclude_classes:
            classes = [cls for cls in classes if cls not in self.inputs.exclude_classes]

        for cls in classes:
            src_dict = src.meta_ext.get_class_dict(cls)
            dest_dict = dest.meta_ext.get_class_dict(cls)
            dest_dict.update(src_dict)
        # Update the shape and slice dimension to reflect the meta extension
        # update.
        dest.meta_ext.slice_dim = src.meta_ext.slice_dim
        dest.meta_ext.shape = src.meta_ext.shape

        self.out_path = op.join(os.getcwd(), op.basename(self.inputs.dest_file))
        dest.to_filename(self.out_path)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["dest_file"] = self.out_path
        return outputs


class MergeNiftiInputSpec(NiftiGeneratorBaseInputSpec):
    in_files = traits.List(mandatory=True, desc="List of Nifti files to merge")
    sort_order = traits.Either(
        traits.Str(),
        traits.List(),
        desc="One or more meta data keys to " "sort files by.",
    )
    merge_dim = traits.Int(
        desc="Dimension to merge along. If not "
        "specified, the last singular or "
        "non-existant dimension is used."
    )


class MergeNiftiOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="Merged Nifti file")


def make_key_func(meta_keys, index=None):
    def key_func(src_nii):
        result = [src_nii.get_meta(key, index) for key in meta_keys]
        return result

    return key_func


class MergeNifti(NiftiGeneratorBase):
    """Merge multiple Nifti files into one. Merges together meta data
    extensions as well."""

    input_spec = MergeNiftiInputSpec
    output_spec = MergeNiftiOutputSpec

    def _run_interface(self, runtime):
        niis = [nb.load(fn) for fn in self.inputs.in_files]
        nws = [NiftiWrapper(nii, make_empty=True) for nii in niis]
        if self.inputs.sort_order:
            sort_order = self.inputs.sort_order
            if isinstance(sort_order, (str, bytes)):
                sort_order = [sort_order]
            nws.sort(key=make_key_func(sort_order))
        if self.inputs.merge_dim == traits.Undefined:
            merge_dim = None
        else:
            merge_dim = self.inputs.merge_dim
        merged = NiftiWrapper.from_sequence(nws, merge_dim)
        const_meta = merged.meta_ext.get_class_dict(("global", "const"))
        self.out_path = self._get_out_path(const_meta)
        nb.save(merged.nii_img, self.out_path)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = self.out_path
        return outputs


class SplitNiftiInputSpec(NiftiGeneratorBaseInputSpec):
    in_file = File(exists=True, mandatory=True, desc="Nifti file to split")
    split_dim = traits.Int(
        desc="Dimension to split along. If not "
        "specified, the last dimension is used."
    )


class SplitNiftiOutputSpec(TraitedSpec):
    out_list = traits.List(File(exists=True), desc="Split Nifti files")


class SplitNifti(NiftiGeneratorBase):
    """
    Split one Nifti file into many along the specified dimension. Each
    result has an updated meta data extension as well.
    """

    input_spec = SplitNiftiInputSpec
    output_spec = SplitNiftiOutputSpec

    def _run_interface(self, runtime):
        self.out_list = []
        nii = nb.load(self.inputs.in_file)
        nw = NiftiWrapper(nii, make_empty=True)
        split_dim = None
        if self.inputs.split_dim == traits.Undefined:
            split_dim = None
        else:
            split_dim = self.inputs.split_dim
        for split_idx, split_nw in enumerate(nw.split(split_dim)):
            const_meta = split_nw.meta_ext.get_class_dict(("global", "const"))
            out_path = self._get_out_path(const_meta, idx=split_idx)
            nb.save(split_nw.nii_img, out_path)
            self.out_list.append(out_path)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_list"] = self.out_list
        return outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..dti import DTIFit


def test_DTIFit_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        base_name=dict(
            argstr="-o %s",
            position=1,
            usedefault=True,
        ),
        bvals=dict(
            argstr="-b %s",
            extensions=None,
            mandatory=True,
            position=4,
        ),
        bvecs=dict(
            argstr="-r %s",
            extensions=None,
            mandatory=True,
            position=3,
        ),
        cni=dict(
            argstr="--cni=%s",
            extensions=None,
        ),
        dwi=dict(
            argstr="-k %s",
            extensions=None,
            mandatory=True,
            position=0,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        gradnonlin=dict(
            argstr="--gradnonlin=%s",
            extensions=None,
        ),
        little_bit=dict(
            argstr="--littlebit",
        ),
        mask=dict(
            argstr="-m %s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
        max_x=dict(
            argstr="-X %d",
        ),
        max_y=dict(
            argstr="-Y %d",
        ),
        max_z=dict(
            argstr="-Z %d",
        ),
        min_x=dict(
            argstr="-x %d",
        ),
        min_y=dict(
            argstr="-y %d",
        ),
        min_z=dict(
            argstr="-z %d",
        ),
        output_type=dict(),
        save_tensor=dict(
            argstr="--save_tensor",
        ),
        sse=dict(
            argstr="--sse",
        ),
    )
    inputs = DTIFit.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DTIFit_outputs():
    output_map = dict(
        FA=dict(
            extensions=None,
        ),
        L1=dict(
            extensions=None,
        ),
        L2=dict(
            extensions=None,
        ),
        L3=dict(
            extensions=None,
        ),
        MD=dict(
            extensions=None,
        ),
        MO=dict(
            extensions=None,
        ),
        S0=dict(
            extensions=None,
        ),
        V1=dict(
            extensions=None,
        ),
        V2=dict(
            extensions=None,
        ),
        V3=dict(
            extensions=None,
        ),
        sse=dict(
            extensions=None,
        ),
        tensor=dict(
            extensions=None,
        ),
    )
    outputs = DTIFit.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# coding: utf-8
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The :py:mod:`nipype.interfaces.elastix` provides the interface to
the elastix registration software.

.. note:: http://elastix.isi.uu.nl/


"""

from ... import logging
from ..base import CommandLineInputSpec, Directory, traits

iflogger = logging.getLogger("nipype.interface")


class ElastixBaseInputSpec(CommandLineInputSpec):
    output_path = Directory(
        "./",
        exists=True,
        mandatory=True,
        usedefault=True,
        argstr="-out %s",
        desc="output directory",
    )
    num_threads = traits.Int(
        1,
        usedefault=True,
        argstr="-threads %01d",
        nohash=True,
        desc="set the maximum number of threads of elastix",
    )



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)
import os


class EMSegmentTransformToNewFormatInputSpec(CommandLineInputSpec):
    inputMRMLFileName = File(
        desc="Active MRML scene that contains EMSegment algorithm parameters in the format before 3.6.3 - please include absolute  file name in path.",
        exists=True,
        argstr="--inputMRMLFileName %s",
    )
    outputMRMLFileName = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Write out the MRML scene after transformation to format 3.6.3 has been made. - has to be in the same directory as the input MRML file due to Slicer Core bug  - please include absolute  file name in path ",
        argstr="--outputMRMLFileName %s",
    )
    templateFlag = traits.Bool(
        desc="Set to true if the transformed mrml file should be used as template file ",
        argstr="--templateFlag ",
    )


class EMSegmentTransformToNewFormatOutputSpec(TraitedSpec):
    outputMRMLFileName = File(
        desc="Write out the MRML scene after transformation to format 3.6.3 has been made. - has to be in the same directory as the input MRML file due to Slicer Core bug  - please include absolute  file name in path ",
        exists=True,
    )


class EMSegmentTransformToNewFormat(SEMLikeCommandLine):
    """title:
  Transform MRML Files to New EMSegmenter Standard


category:
  Utilities


description:
  Transform MRML Files to New EMSegmenter Standard


"""

    input_spec = EMSegmentTransformToNewFormatInputSpec
    output_spec = EMSegmentTransformToNewFormatOutputSpec
    _cmd = "EMSegmentTransformToNewFormat "
    _outputs_filenames = {"outputMRMLFileName": "outputMRMLFileName.mrml"}



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The fsl module provides classes for interfacing with the `FSL
<http://www.fmrib.ox.ac.uk/fsl/index.html>`_ command line tools.  This
was written to work with FSL version 4.1.4.
"""
import os
import warnings

from ...utils.filemanip import fname_presuffix, split_filename, copyfile
from ..base import (
    TraitedSpec,
    isdefined,
    File,
    Directory,
    InputMultiPath,
    OutputMultiPath,
    traits,
)
from .base import FSLCommand, FSLCommandInputSpec, Info


class DTIFitInputSpec(FSLCommandInputSpec):
    dwi = File(
        exists=True,
        desc="diffusion weighted image data file",
        argstr="-k %s",
        position=0,
        mandatory=True,
    )
    base_name = traits.Str(
        "dtifit_",
        desc=("base_name that all output files " "will start with"),
        argstr="-o %s",
        position=1,
        usedefault=True,
    )
    mask = File(
        exists=True,
        desc="bet binary mask file",
        argstr="-m %s",
        position=2,
        mandatory=True,
    )
    bvecs = File(
        exists=True, desc="b vectors file", argstr="-r %s", position=3, mandatory=True
    )
    bvals = File(
        exists=True, desc="b values file", argstr="-b %s", position=4, mandatory=True
    )
    min_z = traits.Int(argstr="-z %d", desc="min z")
    max_z = traits.Int(argstr="-Z %d", desc="max z")
    min_y = traits.Int(argstr="-y %d", desc="min y")
    max_y = traits.Int(argstr="-Y %d", desc="max y")
    min_x = traits.Int(argstr="-x %d", desc="min x")
    max_x = traits.Int(argstr="-X %d", desc="max x")
    save_tensor = traits.Bool(
        desc="save the elements of the tensor", argstr="--save_tensor"
    )
    sse = traits.Bool(desc="output sum of squared errors", argstr="--sse")
    cni = File(exists=True, desc="input counfound regressors", argstr="--cni=%s")
    little_bit = traits.Bool(
        desc="only process small area of brain", argstr="--littlebit"
    )
    gradnonlin = File(
        exists=True, argstr="--gradnonlin=%s", desc="gradient non linearities"
    )


class DTIFitOutputSpec(TraitedSpec):
    V1 = File(exists=True, desc="path/name of file with the 1st eigenvector")
    V2 = File(exists=True, desc="path/name of file with the 2nd eigenvector")
    V3 = File(exists=True, desc="path/name of file with the 3rd eigenvector")
    L1 = File(exists=True, desc="path/name of file with the 1st eigenvalue")
    L2 = File(exists=True, desc="path/name of file with the 2nd eigenvalue")
    L3 = File(exists=True, desc="path/name of file with the 3rd eigenvalue")
    MD = File(exists=True, desc="path/name of file with the mean diffusivity")
    FA = File(exists=True, desc="path/name of file with the fractional anisotropy")
    MO = File(exists=True, desc="path/name of file with the mode of anisotropy")
    S0 = File(
        exists=True,
        desc=(
            "path/name of file with the raw T2 signal with no " "diffusion weighting"
        ),
    )
    tensor = File(exists=True, desc="path/name of file with the 4D tensor volume")
    sse = File(exists=True, desc="path/name of file with the summed squared error")


class DTIFit(FSLCommand):
    """ Use FSL  dtifit command for fitting a diffusion tensor model at each
    voxel

    Example
    -------

    >>> from nipype.interfaces import fsl
    >>> dti = fsl.DTIFit()
    >>> dti.inputs.dwi = 'diffusion.nii'
    >>> dti.inputs.bvecs = 'bvecs'
    >>> dti.inputs.bvals = 'bvals'
    >>> dti.inputs.base_name = 'TP'
    >>> dti.inputs.mask = 'mask.nii'
    >>> dti.cmdline
    'dtifit -k diffusion.nii -o TP -m mask.nii -r bvecs -b bvals'

    """

    _cmd = "dtifit"
    input_spec = DTIFitInputSpec
    output_spec = DTIFitOutputSpec

    def _list_outputs(self):
        keys_to_ignore = {"outputtype", "environ", "args"}
        # Optional output: Map output name to input flag
        opt_output = {"tensor": self.inputs.save_tensor, "sse": self.inputs.sse}
        # Ignore optional output, whose corresponding input-flag is not defined
        # or set to False
        for output, input_flag in opt_output.items():
            if isdefined(input_flag) and input_flag:
                # this is wanted output, do not ignore
                continue
            keys_to_ignore.add(output)

        outputs = self.output_spec().get()
        for k in set(outputs.keys()) - keys_to_ignore:
            outputs[k] = self._gen_fname(self.inputs.base_name, suffix="_" + k)
        return outputs


class FSLXCommandInputSpec(FSLCommandInputSpec):
    dwi = File(
        exists=True,
        argstr="--data=%s",
        mandatory=True,
        desc="diffusion weighted image data file",
    )
    mask = File(
        exists=True,
        argstr="--mask=%s",
        mandatory=True,
        desc="brain binary mask file (i.e. from BET)",
    )
    bvecs = File(
        exists=True, argstr="--bvecs=%s", mandatory=True, desc="b vectors file"
    )
    bvals = File(exists=True, argstr="--bvals=%s", mandatory=True, desc="b values file")

    logdir = Directory(".", argstr="--logdir=%s", usedefault=True)
    n_fibres = traits.Range(
        usedefault=True,
        low=1,
        value=2,
        argstr="--nfibres=%d",
        desc=("Maximum number of fibres to fit in each voxel"),
        mandatory=True,
    )
    model = traits.Enum(
        1,
        2,
        3,
        argstr="--model=%d",
        desc=(
            "use monoexponential (1, default, required for "
            "single-shell) or multiexponential (2, multi-"
            "shell) model"
        ),
    )
    fudge = traits.Int(argstr="--fudge=%d", desc="ARD fudge factor")
    n_jumps = traits.Int(
        5000,
        usedefault=True,
        argstr="--njumps=%d",
        desc="Num of jumps to be made by MCMC",
    )
    burn_in = traits.Range(
        low=0,
        value=0,
        usedefault=True,
        argstr="--burnin=%d",
        desc=("Total num of jumps at start of MCMC to be " "discarded"),
    )
    burn_in_no_ard = traits.Range(
        low=0,
        value=0,
        usedefault=True,
        argstr="--burnin_noard=%d",
        desc=("num of burnin jumps before the ard is" " imposed"),
    )
    sample_every = traits.Range(
        low=0,
        value=1,
        usedefault=True,
        argstr="--sampleevery=%d",
        desc="Num of jumps for each sample (MCMC)",
    )
    update_proposal_every = traits.Range(
        low=1,
        value=40,
        usedefault=True,
        argstr="--updateproposalevery=%d",
        desc=("Num of jumps for each update " "to the proposal density std " "(MCMC)"),
    )
    seed = traits.Int(
        argstr="--seed=%d", desc="seed for pseudo random number generator"
    )

    _xor_inputs1 = ("no_ard", "all_ard")
    no_ard = traits.Bool(
        argstr="--noard", xor=_xor_inputs1, desc="Turn ARD off on all fibres"
    )
    all_ard = traits.Bool(
        argstr="--allard", xor=_xor_inputs1, desc="Turn ARD on on all fibres"
    )

    _xor_inputs2 = ("no_spat", "non_linear", "cnlinear")
    no_spat = traits.Bool(
        argstr="--nospat",
        xor=_xor_inputs2,
        desc="Initialise with tensor, not spatially",
    )
    non_linear = traits.Bool(
        argstr="--nonlinear", xor=_xor_inputs2, desc="Initialise with nonlinear fitting"
    )
    cnlinear = traits.Bool(
        argstr="--cnonlinear",
        xor=_xor_inputs2,
        desc=("Initialise with constrained nonlinear " "fitting"),
    )
    rician = traits.Bool(argstr="--rician", desc=("use Rician noise modeling"))

    _xor_inputs3 = ["f0_noard", "f0_ard"]
    f0_noard = traits.Bool(
        argstr="--f0",
        xor=_xor_inputs3,
        desc=(
            "Noise floor model: add to the model an "
            "unattenuated signal compartment f0"
        ),
    )
    f0_ard = traits.Bool(
        argstr="--f0 --ardf0",
        xor=_xor_inputs3 + ["all_ard"],
        desc=(
            "Noise floor model: add to the model an "
            "unattenuated signal compartment f0"
        ),
    )
    force_dir = traits.Bool(
        True,
        argstr="--forcedir",
        usedefault=True,
        desc=(
            "use the actual directory name given "
            "(do not add + to make a new directory)"
        ),
    )


class FSLXCommandOutputSpec(TraitedSpec):
    dyads = OutputMultiPath(
        File(exists=True), desc=("Mean of PDD distribution" " in vector form.")
    )
    fsamples = OutputMultiPath(
        File(exists=True), desc=("Samples from the " "distribution on f " "anisotropy")
    )
    mean_dsamples = File(exists=True, desc="Mean of distribution on diffusivity d")
    mean_fsamples = OutputMultiPath(
        File(exists=True), desc=("Mean of distribution on f " "anisotropy")
    )
    mean_S0samples = File(
        exists=True, desc=("Mean of distribution on T2w" "baseline signal intensity S0")
    )
    mean_tausamples = File(
        exists=True,
        desc=("Mean of distribution on " "tau samples (only with rician " "noise)"),
    )
    phsamples = OutputMultiPath(File(exists=True), desc=("phi samples, per fiber"))
    thsamples = OutputMultiPath(File(exists=True), desc=("theta samples, per fiber"))


class FSLXCommand(FSLCommand):
    """
    Base support for ``xfibres`` and ``bedpostx``
    """

    input_spec = FSLXCommandInputSpec
    output_spec = FSLXCommandOutputSpec

    def _run_interface(self, runtime):
        self._out_dir = os.getcwd()
        runtime = super(FSLXCommand, self)._run_interface(runtime)
        if runtime.stderr:
            self.raise_exception(runtime)
        return runtime

    def _list_outputs(self, out_dir=None):
        outputs = self.output_spec().get()
        n_fibres = self.inputs.n_fibres
        if not out_dir:
            if isdefined(self.inputs.logdir):
                out_dir = os.path.abspath(self.inputs.logdir)
            else:
                out_dir = os.path.abspath("logdir")

        multi_out = ["dyads", "fsamples", "mean_fsamples", "phsamples", "thsamples"]
        single_out = ["mean_dsamples", "mean_S0samples"]

        for k in single_out:
            outputs[k] = self._gen_fname(k, cwd=out_dir)

        if isdefined(self.inputs.rician) and self.inputs.rician:
            outputs["mean_tausamples"] = self._gen_fname("mean_tausamples", cwd=out_dir)

        for k in multi_out:
            outputs[k] = []

        for i in range(1, n_fibres + 1):
            outputs["fsamples"].append(self._gen_fname("f%dsamples" % i, cwd=out_dir))
            outputs["mean_fsamples"].append(
                self._gen_fname("mean_f%dsamples" % i, cwd=out_dir)
            )

        for i in range(1, n_fibres + 1):
            outputs["dyads"].append(self._gen_fname("dyads%d" % i, cwd=out_dir))
            outputs["phsamples"].append(self._gen_fname("ph%dsamples" % i, cwd=out_dir))
            outputs["thsamples"].append(self._gen_fname("th%dsamples" % i, cwd=out_dir))

        return outputs


class BEDPOSTX5InputSpec(FSLXCommandInputSpec):
    dwi = File(exists=True, desc="diffusion weighted image data file", mandatory=True)
    mask = File(exists=True, desc="bet binary mask file", mandatory=True)
    bvecs = File(exists=True, desc="b vectors file", mandatory=True)
    bvals = File(exists=True, desc="b values file", mandatory=True)
    logdir = Directory(argstr="--logdir=%s")
    n_fibres = traits.Range(
        usedefault=True,
        low=1,
        value=2,
        argstr="-n %d",
        desc=("Maximum number of fibres to fit in each voxel"),
        mandatory=True,
    )
    model = traits.Enum(
        1,
        2,
        3,
        argstr="-model %d",
        desc=(
            "use monoexponential (1, default, required for "
            "single-shell) or multiexponential (2, multi-"
            "shell) model"
        ),
    )
    fudge = traits.Int(argstr="-w %d", desc="ARD fudge factor")
    n_jumps = traits.Int(
        5000, usedefault=True, argstr="-j %d", desc="Num of jumps to be made by MCMC"
    )
    burn_in = traits.Range(
        low=0,
        value=0,
        usedefault=True,
        argstr="-b %d",
        desc=("Total num of jumps at start of MCMC to be " "discarded"),
    )
    sample_every = traits.Range(
        low=0,
        value=1,
        usedefault=True,
        argstr="-s %d",
        desc="Num of jumps for each sample (MCMC)",
    )
    out_dir = Directory(
        "bedpostx",
        mandatory=True,
        desc="output directory",
        usedefault=True,
        position=1,
        argstr="%s",
    )
    gradnonlin = traits.Bool(
        False, argstr="-g", desc=("consider gradient nonlinearities, " "default off")
    )
    grad_dev = File(exists=True, desc="grad_dev file, if gradnonlin, -g is True")
    use_gpu = traits.Bool(False, desc="Use the GPU version of bedpostx")


class BEDPOSTX5OutputSpec(TraitedSpec):
    mean_dsamples = File(exists=True, desc="Mean of distribution on diffusivity d")
    mean_fsamples = OutputMultiPath(
        File(exists=True), desc=("Mean of distribution on f " "anisotropy")
    )
    mean_S0samples = File(
        exists=True, desc=("Mean of distribution on T2w" "baseline signal intensity S0")
    )
    mean_phsamples = OutputMultiPath(
        File(exists=True), desc="Mean of distribution on phi"
    )
    mean_thsamples = OutputMultiPath(
        File(exists=True), desc="Mean of distribution on theta"
    )
    merged_thsamples = OutputMultiPath(
        File(exists=True), desc=("Samples from the distribution " "on theta")
    )
    merged_phsamples = OutputMultiPath(
        File(exists=True), desc=("Samples from the distribution " "on phi")
    )
    merged_fsamples = OutputMultiPath(
        File(exists=True),
        desc=("Samples from the distribution on " "anisotropic volume fraction"),
    )
    dyads = OutputMultiPath(
        File(exists=True), desc="Mean of PDD distribution in vector form."
    )
    dyads_dispersion = OutputMultiPath(File(exists=True), desc=("Dispersion"))


class BEDPOSTX5(FSLXCommand):
    """
    BEDPOSTX stands for Bayesian Estimation of Diffusion Parameters Obtained
    using Sampling Techniques. The X stands for modelling Crossing Fibres.
    bedpostx runs Markov Chain Monte Carlo sampling to build up distributions
    on diffusion parameters at each voxel. It creates all the files necessary
    for running probabilistic tractography. For an overview of the modelling
    carried out within bedpostx see this `technical report
    <http://www.fmrib.ox.ac.uk/analysis/techrep/tr03tb1/tr03tb1/index.html>`_.


    .. note:: Consider using
      :func:`niflow.nipype1.workflows.fsl.dmri.create_bedpostx_pipeline` instead.


    Example
    -------

    >>> from nipype.interfaces import fsl
    >>> bedp = fsl.BEDPOSTX5(bvecs='bvecs', bvals='bvals', dwi='diffusion.nii',
    ...                     mask='mask.nii', n_fibres=1)
    >>> bedp.cmdline
    'bedpostx bedpostx -b 0 --burnin_noard=0 --forcedir -n 1 -j 5000 \
-s 1 --updateproposalevery=40'

    """

    _cmd = "bedpostx"
    _default_cmd = _cmd
    input_spec = BEDPOSTX5InputSpec
    output_spec = BEDPOSTX5OutputSpec
    _can_resume = True

    def __init__(self, **inputs):
        super(BEDPOSTX5, self).__init__(**inputs)
        self.inputs.on_trait_change(self._cuda_update, "use_gpu")

    def _cuda_update(self):
        if isdefined(self.inputs.use_gpu) and self.inputs.use_gpu:
            self._cmd = "bedpostx_gpu"
        else:
            self._cmd = self._default_cmd

    def _run_interface(self, runtime):

        subjectdir = os.path.abspath(self.inputs.out_dir)
        if not os.path.exists(subjectdir):
            os.makedirs(subjectdir)
        _, _, ext = split_filename(self.inputs.mask)
        copyfile(self.inputs.mask, os.path.join(subjectdir, "nodif_brain_mask" + ext))
        _, _, ext = split_filename(self.inputs.dwi)
        copyfile(self.inputs.dwi, os.path.join(subjectdir, "data" + ext))
        copyfile(self.inputs.bvals, os.path.join(subjectdir, "bvals"))
        copyfile(self.inputs.bvecs, os.path.join(subjectdir, "bvecs"))
        if isdefined(self.inputs.grad_dev):
            _, _, ext = split_filename(self.inputs.grad_dev)
            copyfile(self.inputs.grad_dev, os.path.join(subjectdir, "grad_dev" + ext))

        retval = super(BEDPOSTX5, self)._run_interface(runtime)

        self._out_dir = subjectdir + ".bedpostX"
        return retval

    def _list_outputs(self):
        outputs = self.output_spec().get()
        n_fibres = self.inputs.n_fibres

        multi_out = [
            "merged_thsamples",
            "merged_fsamples",
            "merged_phsamples",
            "mean_phsamples",
            "mean_thsamples",
            "mean_fsamples",
            "dyads_dispersion",
            "dyads",
        ]

        single_out = ["mean_dsamples", "mean_S0samples"]

        for k in single_out:
            outputs[k] = self._gen_fname(k, cwd=self._out_dir)

        for k in multi_out:
            outputs[k] = []

        for i in range(1, n_fibres + 1):
            outputs["merged_thsamples"].append(
                self._gen_fname("merged_th%dsamples" % i, cwd=self._out_dir)
            )
            outputs["merged_fsamples"].append(
                self._gen_fname("merged_f%dsamples" % i, cwd=self._out_dir)
            )
            outputs["merged_phsamples"].append(
                self._gen_fname("merged_ph%dsamples" % i, cwd=self._out_dir)
            )
            outputs["mean_thsamples"].append(
                self._gen_fname("mean_th%dsamples" % i, cwd=self._out_dir)
            )
            outputs["mean_phsamples"].append(
                self._gen_fname("mean_ph%dsamples" % i, cwd=self._out_dir)
            )
            outputs["mean_fsamples"].append(
                self._gen_fname("mean_f%dsamples" % i, cwd=self._out_dir)
            )
            outputs["dyads"].append(self._gen_fname("dyads%d" % i, cwd=self._out_dir))
            outputs["dyads_dispersion"].append(
                self._gen_fname("dyads%d_dispersion" % i, cwd=self._out_dir)
            )
        return outputs


class XFibres5InputSpec(FSLXCommandInputSpec):
    gradnonlin = File(
        exists=True,
        argstr="--gradnonlin=%s",
        desc="gradient file corresponding to slice",
    )


class XFibres5(FSLXCommand):
    """
    Perform model parameters estimation for local (voxelwise) diffusion
    parameters
    """

    _cmd = "xfibres"
    input_spec = XFibres5InputSpec
    output_spec = FSLXCommandOutputSpec


XFibres = XFibres5
BEDPOSTX = BEDPOSTX5


class ProbTrackXBaseInputSpec(FSLCommandInputSpec):
    thsamples = InputMultiPath(File(exists=True), mandatory=True)
    phsamples = InputMultiPath(File(exists=True), mandatory=True)
    fsamples = InputMultiPath(File(exists=True), mandatory=True)
    samples_base_name = traits.Str(
        "merged",
        desc=("the rootname/base_name for samples " "files"),
        argstr="--samples=%s",
        usedefault=True,
    )
    mask = File(
        exists=True,
        desc="bet binary mask file in diffusion space",
        argstr="-m %s",
        mandatory=True,
    )
    seed = traits.Either(
        File(exists=True),
        traits.List(File(exists=True)),
        traits.List(traits.List(traits.Int(), minlen=3, maxlen=3)),
        desc=("seed volume(s), or voxel(s) or freesurfer " "label file"),
        argstr="--seed=%s",
        mandatory=True,
    )
    target_masks = InputMultiPath(
        File(exits=True),
        desc=("list of target masks - required for " "seeds_to_targets classification"),
        argstr="--targetmasks=%s",
    )
    waypoints = File(
        exists=True,
        desc=(
            "waypoint mask or ascii list of waypoint masks - "
            "only keep paths going through ALL the masks"
        ),
        argstr="--waypoints=%s",
    )
    network = traits.Bool(
        desc=(
            "activate network mode - only keep paths "
            "going through at least one seed mask "
            "(required if multiple seed masks)"
        ),
        argstr="--network",
    )
    seed_ref = File(
        exists=True,
        desc=(
            "reference vol to define seed space in simple mode "
            "- diffusion space assumed if absent"
        ),
        argstr="--seedref=%s",
    )
    out_dir = Directory(
        exists=True,
        argstr="--dir=%s",
        desc="directory to put the final volumes in",
        genfile=True,
    )
    force_dir = traits.Bool(
        True,
        desc=(
            "use the actual directory name given - i.e. "
            "do not add + to make a new directory"
        ),
        argstr="--forcedir",
        usedefault=True,
    )
    opd = traits.Bool(
        True, desc="outputs path distributions", argstr="--opd", usedefault=True
    )
    correct_path_distribution = traits.Bool(
        desc=("correct path distribution " "for the length of the " "pathways"),
        argstr="--pd",
    )
    os2t = traits.Bool(desc="Outputs seeds to targets", argstr="--os2t")
    # paths_file = File('nipype_fdtpaths', usedefault=True, argstr='--out=%s',
    #                 desc='produces an output file (default is fdt_paths)')
    avoid_mp = File(
        exists=True,
        desc=("reject pathways passing through locations given by " "this mask"),
        argstr="--avoid=%s",
    )
    stop_mask = File(
        exists=True,
        argstr="--stop=%s",
        desc="stop tracking at locations given by this mask file",
    )
    xfm = File(
        exists=True,
        argstr="--xfm=%s",
        desc=(
            "transformation matrix taking seed space to DTI space "
            "(either FLIRT matrix or FNIRT warp_field) - default is "
            "identity"
        ),
    )
    inv_xfm = File(
        argstr="--invxfm=%s",
        desc=(
            "transformation matrix taking DTI space to seed "
            "space (compulsory when using a warp_field for "
            "seeds_to_dti)"
        ),
    )
    n_samples = traits.Int(
        5000,
        argstr="--nsamples=%d",
        desc="number of samples - default=5000",
        usedefault=True,
    )
    n_steps = traits.Int(
        argstr="--nsteps=%d", desc="number of steps per sample - default=2000"
    )
    dist_thresh = traits.Float(
        argstr="--distthresh=%.3f",
        desc=("discards samples shorter than this " "threshold (in mm - default=0)"),
    )
    c_thresh = traits.Float(
        argstr="--cthr=%.3f", desc="curvature threshold - default=0.2"
    )
    sample_random_points = traits.Bool(
        argstr="--sampvox", desc=("sample random points within " "seed voxels")
    )
    step_length = traits.Float(
        argstr="--steplength=%.3f", desc="step_length in mm - default=0.5"
    )
    loop_check = traits.Bool(
        argstr="--loopcheck",
        desc=(
            "perform loop_checks on paths - slower, "
            "but allows lower curvature threshold"
        ),
    )
    use_anisotropy = traits.Bool(
        argstr="--usef", desc="use anisotropy to constrain tracking"
    )
    rand_fib = traits.Enum(
        0,
        1,
        2,
        3,
        argstr="--randfib=%d",
        desc=(
            "options: 0 - default, 1 - to randomly "
            "sample initial fibres (with f > fibthresh), "
            "2 - to sample in proportion fibres (with "
            "f>fibthresh) to f, 3 - to sample ALL "
            "populations at random (even if "
            "f<fibthresh)"
        ),
    )
    fibst = traits.Int(
        argstr="--fibst=%d",
        desc=(
            "force a starting fibre for tracking - "
            "default=1, i.e. first fibre orientation. Only "
            "works if randfib==0"
        ),
    )
    mod_euler = traits.Bool(argstr="--modeuler", desc="use modified euler streamlining")
    random_seed = traits.Bool(argstr="--rseed", desc="random seed")
    s2tastext = traits.Bool(
        argstr="--s2tastext",
        desc=(
            "output seed-to-target counts as a text "
            "file (useful when seeding from a mesh)"
        ),
    )
    verbose = traits.Enum(
        0,
        1,
        2,
        desc=("Verbose level, [0-2]. Level 2 is required to " "output particle files."),
        argstr="--verbose=%d",
    )


class ProbTrackXInputSpec(ProbTrackXBaseInputSpec):
    mode = traits.Enum(
        "simple",
        "two_mask_symm",
        "seedmask",
        desc=(
            "options: simple (single seed voxel), seedmask "
            "(mask of seed voxels), twomask_symm (two bet "
            "binary masks)"
        ),
        argstr="--mode=%s",
        genfile=True,
    )
    mask2 = File(
        exists=True,
        desc=("second bet binary mask (in diffusion space) in " "twomask_symm mode"),
        argstr="--mask2=%s",
    )
    mesh = File(
        exists=True,
        desc="Freesurfer-type surface descriptor (in ascii format)",
        argstr="--mesh=%s",
    )


class ProbTrackXOutputSpec(TraitedSpec):
    log = File(
        exists=True, desc="path/name of a text record of the command that was run"
    )
    fdt_paths = OutputMultiPath(
        File(exists=True),
        desc=(
            "path/name of a 3D image file "
            "containing the output connectivity "
            "distribution to the seed mask"
        ),
    )
    way_total = File(
        exists=True,
        desc=(
            "path/name of a text file containing a single "
            "number corresponding to the total number of "
            "generated tracts that have not been rejected by "
            "inclusion/exclusion mask criteria"
        ),
    )
    targets = traits.List(
        File(exists=True), desc=("a list with all generated seeds_to_target " "files")
    )
    particle_files = traits.List(
        File(exists=True),
        desc=(
            "Files describing all of the tract "
            "samples. Generated only if verbose is "
            "set to 2"
        ),
    )


class ProbTrackX(FSLCommand):
    """ Use FSL  probtrackx for tractography on bedpostx results

    Examples
    --------

    >>> from nipype.interfaces import fsl
    >>> pbx = fsl.ProbTrackX(samples_base_name='merged', mask='mask.nii', \
    seed='MASK_average_thal_right.nii', mode='seedmask', \
    xfm='trans.mat', n_samples=3, n_steps=10, force_dir=True, opd=True, \
    os2t=True, target_masks = ['targets_MASK1.nii', 'targets_MASK2.nii'], \
    thsamples='merged_thsamples.nii', fsamples='merged_fsamples.nii', \
    phsamples='merged_phsamples.nii', out_dir='.')
    >>> pbx.cmdline
    'probtrackx --forcedir -m mask.nii --mode=seedmask --nsamples=3 --nsteps=10 --opd --os2t --dir=. --samples=merged --seed=MASK_average_thal_right.nii --targetmasks=targets.txt --xfm=trans.mat'

    """

    _cmd = "probtrackx"
    input_spec = ProbTrackXInputSpec
    output_spec = ProbTrackXOutputSpec

    def __init__(self, **inputs):
        warnings.warn(
            ("Deprecated: Please use create_bedpostx_pipeline " "instead"),
            DeprecationWarning,
        )
        return super(ProbTrackX, self).__init__(**inputs)

    def _run_interface(self, runtime):
        for i in range(1, len(self.inputs.thsamples) + 1):
            _, _, ext = split_filename(self.inputs.thsamples[i - 1])
            copyfile(
                self.inputs.thsamples[i - 1],
                self.inputs.samples_base_name + "_th%dsamples" % i + ext,
                copy=False,
            )
            _, _, ext = split_filename(self.inputs.thsamples[i - 1])
            copyfile(
                self.inputs.phsamples[i - 1],
                self.inputs.samples_base_name + "_ph%dsamples" % i + ext,
                copy=False,
            )
            _, _, ext = split_filename(self.inputs.thsamples[i - 1])
            copyfile(
                self.inputs.fsamples[i - 1],
                self.inputs.samples_base_name + "_f%dsamples" % i + ext,
                copy=False,
            )

        if isdefined(self.inputs.target_masks):
            f = open("targets.txt", "w")
            for target in self.inputs.target_masks:
                f.write("%s\n" % target)
            f.close()
        if isinstance(self.inputs.seed, list):
            f = open("seeds.txt", "w")
            for seed in self.inputs.seed:
                if isinstance(seed, list):
                    f.write("%s\n" % (" ".join([str(s) for s in seed])))
                else:
                    f.write("%s\n" % seed)
            f.close()

        runtime = super(ProbTrackX, self)._run_interface(runtime)
        if runtime.stderr:
            self.raise_exception(runtime)
        return runtime

    def _format_arg(self, name, spec, value):
        if name == "target_masks" and isdefined(value):
            fname = "targets.txt"
            return super(ProbTrackX, self)._format_arg(name, spec, [fname])
        elif name == "seed" and isinstance(value, list):
            fname = "seeds.txt"
            return super(ProbTrackX, self)._format_arg(name, spec, fname)
        else:
            return super(ProbTrackX, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        if not isdefined(self.inputs.out_dir):
            out_dir = self._gen_filename("out_dir")
        else:
            out_dir = self.inputs.out_dir

        outputs["log"] = os.path.abspath(os.path.join(out_dir, "probtrackx.log"))
        # utputs['way_total'] = os.path.abspath(os.path.join(out_dir,
        #                                                    'waytotal'))
        if isdefined(self.inputs.opd is True):
            if isinstance(self.inputs.seed, list) and isinstance(
                self.inputs.seed[0], list
            ):
                outputs["fdt_paths"] = []
                for seed in self.inputs.seed:
                    outputs["fdt_paths"].append(
                        os.path.abspath(
                            self._gen_fname(
                                ("fdt_paths_%s" % ("_".join([str(s) for s in seed]))),
                                cwd=out_dir,
                                suffix="",
                            )
                        )
                    )
            else:
                outputs["fdt_paths"] = os.path.abspath(
                    self._gen_fname("fdt_paths", cwd=out_dir, suffix="")
                )

        # handle seeds-to-target output files
        if isdefined(self.inputs.target_masks):
            outputs["targets"] = []
            for target in self.inputs.target_masks:
                outputs["targets"].append(
                    os.path.abspath(
                        self._gen_fname(
                            "seeds_to_" + os.path.split(target)[1],
                            cwd=out_dir,
                            suffix="",
                        )
                    )
                )
        if isdefined(self.inputs.verbose) and self.inputs.verbose == 2:
            outputs["particle_files"] = [
                os.path.abspath(os.path.join(out_dir, "particle%d" % i))
                for i in range(self.inputs.n_samples)
            ]
        return outputs

    def _gen_filename(self, name):
        if name == "out_dir":
            return os.getcwd()
        elif name == "mode":
            if isinstance(self.inputs.seed, list) and isinstance(
                self.inputs.seed[0], list
            ):
                return "simple"
            else:
                return "seedmask"


class ProbTrackX2InputSpec(ProbTrackXBaseInputSpec):
    simple = traits.Bool(
        desc=(
            "rack from a list of voxels (seed must be a " "ASCII list of coordinates)"
        ),
        argstr="--simple",
    )
    fopd = File(
        exists=True,
        desc="Other mask for binning tract distribution",
        argstr="--fopd=%s",
    )
    waycond = traits.Enum(
        "OR",
        "AND",
        argstr="--waycond=%s",
        desc=('Waypoint condition. Either "AND" (default) ' 'or "OR"'),
    )
    wayorder = traits.Bool(
        desc=(
            "Reject streamlines that do not hit "
            "waypoints in given order. Only valid if "
            "waycond=AND"
        ),
        argstr="--wayorder",
    )
    onewaycondition = traits.Bool(
        desc=("Apply waypoint conditions to each " "half tract separately"),
        argstr="--onewaycondition",
    )
    omatrix1 = traits.Bool(
        desc="Output matrix1 - SeedToSeed Connectivity", argstr="--omatrix1"
    )
    distthresh1 = traits.Float(
        argstr="--distthresh1=%.3f",
        desc=(
            "Discards samples (in matrix1) shorter "
            "than this threshold (in mm - "
            "default=0)"
        ),
    )
    omatrix2 = traits.Bool(
        desc="Output matrix2 - SeedToLowResMask",
        argstr="--omatrix2",
        requires=["target2"],
    )
    target2 = File(
        exists=True,
        desc=(
            "Low resolution binary brain mask for storing "
            "connectivity distribution in matrix2 mode"
        ),
        argstr="--target2=%s",
    )
    omatrix3 = traits.Bool(
        desc="Output matrix3 (NxN connectivity matrix)",
        argstr="--omatrix3",
        requires=["target3", "lrtarget3"],
    )
    target3 = File(
        exists=True,
        desc=("Mask used for NxN connectivity matrix (or Nxn if " "lrtarget3 is set)"),
        argstr="--target3=%s",
    )
    lrtarget3 = File(
        exists=True,
        desc="Column-space mask used for Nxn connectivity matrix",
        argstr="--lrtarget3=%s",
    )
    distthresh3 = traits.Float(
        argstr="--distthresh3=%.3f",
        desc=(
            "Discards samples (in matrix3) shorter "
            "than this threshold (in mm - "
            "default=0)"
        ),
    )
    omatrix4 = traits.Bool(
        desc=("Output matrix4 - DtiMaskToSeed (special " "Oxford Sparse Format)"),
        argstr="--omatrix4",
    )
    colmask4 = File(
        exists=True,
        desc="Mask for columns of matrix4 (default=seed mask)",
        argstr="--colmask4=%s",
    )
    target4 = File(exists=True, desc="Brain mask in DTI space", argstr="--target4=%s")
    meshspace = traits.Enum(
        "caret",
        "freesurfer",
        "first",
        "vox",
        argstr="--meshspace=%s",
        desc=(
            'Mesh reference space - either "caret" '
            '(default) or "freesurfer" or "first" or '
            '"vox"'
        ),
    )


class ProbTrackX2OutputSpec(ProbTrackXOutputSpec):
    network_matrix = File(
        exists=True, desc=("the network matrix generated by --omatrix1 " "option")
    )
    matrix1_dot = File(exists=True, desc="Output matrix1.dot - SeedToSeed Connectivity")
    lookup_tractspace = File(
        exists=True, desc=("lookup_tractspace generated by " "--omatrix2 option")
    )
    matrix2_dot = File(exists=True, desc="Output matrix2.dot - SeedToLowResMask")
    matrix3_dot = File(exists=True, desc="Output matrix3 - NxN connectivity matrix")


class ProbTrackX2(ProbTrackX):
    """ Use FSL  probtrackx2 for tractography on bedpostx results

    Examples
    --------

    >>> from nipype.interfaces import fsl
    >>> pbx2 = fsl.ProbTrackX2()
    >>> pbx2.inputs.seed = 'seed_source.nii.gz'
    >>> pbx2.inputs.thsamples = 'merged_th1samples.nii.gz'
    >>> pbx2.inputs.fsamples = 'merged_f1samples.nii.gz'
    >>> pbx2.inputs.phsamples = 'merged_ph1samples.nii.gz'
    >>> pbx2.inputs.mask = 'nodif_brain_mask.nii.gz'
    >>> pbx2.inputs.out_dir = '.'
    >>> pbx2.inputs.n_samples = 3
    >>> pbx2.inputs.n_steps = 10
    >>> pbx2.cmdline
    'probtrackx2 --forcedir -m nodif_brain_mask.nii.gz --nsamples=3 --nsteps=10 --opd --dir=. --samples=merged --seed=seed_source.nii.gz'
    """

    _cmd = "probtrackx2"
    input_spec = ProbTrackX2InputSpec
    output_spec = ProbTrackX2OutputSpec

    def _list_outputs(self):
        outputs = super(ProbTrackX2, self)._list_outputs()

        if not isdefined(self.inputs.out_dir):
            out_dir = os.getcwd()
        else:
            out_dir = self.inputs.out_dir

        outputs["way_total"] = os.path.abspath(os.path.join(out_dir, "waytotal"))

        if isdefined(self.inputs.omatrix1):
            outputs["network_matrix"] = os.path.abspath(
                os.path.join(out_dir, "matrix_seeds_to_all_targets")
            )
            outputs["matrix1_dot"] = os.path.abspath(
                os.path.join(out_dir, "fdt_matrix1.dot")
            )

        if isdefined(self.inputs.omatrix2):
            outputs["lookup_tractspace"] = os.path.abspath(
                os.path.join(out_dir, "lookup_tractspace_fdt_matrix2.nii.gz")
            )
            outputs["matrix2_dot"] = os.path.abspath(
                os.path.join(out_dir, "fdt_matrix2.dot")
            )

        if isdefined(self.inputs.omatrix3):
            outputs["matrix3_dot"] = os.path.abspath(
                os.path.join(out_dir, "fdt_matrix3.dot")
            )
        return outputs


class VecRegInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        argstr="-i %s",
        desc="filename for input vector or tensor field",
        mandatory=True,
    )
    out_file = File(
        argstr="-o %s",
        desc=("filename for output registered vector or tensor " "field"),
        genfile=True,
        hash_files=False,
    )
    ref_vol = File(
        exists=True,
        argstr="-r %s",
        desc="filename for reference (target) volume",
        mandatory=True,
    )
    affine_mat = File(
        exists=True, argstr="-t %s", desc="filename for affine transformation matrix"
    )
    warp_field = File(
        exists=True,
        argstr="-w %s",
        desc=("filename for 4D warp field for nonlinear " "registration"),
    )
    rotation_mat = File(
        exists=True,
        argstr="--rotmat=%s",
        desc=(
            "filename for secondary affine matrix if set, "
            "this will be used for the rotation of the "
            "vector/tensor field"
        ),
    )
    rotation_warp = File(
        exists=True,
        argstr="--rotwarp=%s",
        desc=(
            "filename for secondary warp field if set, "
            "this will be used for the rotation of the "
            "vector/tensor field"
        ),
    )
    interpolation = traits.Enum(
        "nearestneighbour",
        "trilinear",
        "sinc",
        "spline",
        argstr="--interp=%s",
        desc=(
            "interpolation method : "
            "nearestneighbour, trilinear (default), "
            "sinc or spline"
        ),
    )
    mask = File(exists=True, argstr="-m %s", desc="brain mask in input space")
    ref_mask = File(
        exists=True,
        argstr="--refmask=%s",
        desc=("brain mask in output space (useful for speed up of " "nonlinear reg)"),
    )


class VecRegOutputSpec(TraitedSpec):
    out_file = File(
        exists=True,
        desc=("path/name of filename for the registered vector or " "tensor field"),
    )


class VecReg(FSLCommand):
    """Use FSL vecreg for registering vector data
    For complete details, see the FDT Documentation
    <http://www.fmrib.ox.ac.uk/fsl/fdt/fdt_vecreg.html>

    Example
    -------

    >>> from nipype.interfaces import fsl
    >>> vreg = fsl.VecReg(in_file='diffusion.nii', \
                 affine_mat='trans.mat', \
                 ref_vol='mni.nii', \
                 out_file='diffusion_vreg.nii')
    >>> vreg.cmdline
    'vecreg -t trans.mat -i diffusion.nii -o diffusion_vreg.nii -r mni.nii'

    """

    _cmd = "vecreg"
    input_spec = VecRegInputSpec
    output_spec = VecRegOutputSpec

    def _run_interface(self, runtime):
        if not isdefined(self.inputs.out_file):
            pth, base_name = os.path.split(self.inputs.in_file)
            self.inputs.out_file = self._gen_fname(
                base_name, cwd=os.path.abspath(pth), suffix="_vreg"
            )
        return super(VecReg, self)._run_interface(runtime)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_file
        if not isdefined(outputs["out_file"]) and isdefined(self.inputs.in_file):
            pth, base_name = os.path.split(self.inputs.in_file)
            outputs["out_file"] = self._gen_fname(
                base_name, cwd=os.path.abspath(pth), suffix="_vreg"
            )
        outputs["out_file"] = os.path.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._list_outputs()[name]
        else:
            return None


class ProjThreshInputSpec(FSLCommandInputSpec):
    in_files = traits.List(
        File(exists=True),
        argstr="%s",
        desc="a list of input volumes",
        mandatory=True,
        position=0,
    )
    threshold = traits.Int(
        argstr="%d",
        desc=(
            "threshold indicating minimum number of seed "
            "voxels entering this mask region"
        ),
        mandatory=True,
        position=1,
    )


class ProjThreshOuputSpec(TraitedSpec):
    out_files = traits.List(
        File(exists=True), desc=("path/name of output volume after " "thresholding")
    )


class ProjThresh(FSLCommand):
    """Use FSL proj_thresh for thresholding some outputs of probtrack
    For complete details, see the FDT Documentation
    <http://www.fmrib.ox.ac.uk/fsl/fdt/fdt_thresh.html>

    Example
    -------

    >>> from nipype.interfaces import fsl
    >>> ldir = ['seeds_to_M1.nii', 'seeds_to_M2.nii']
    >>> pThresh = fsl.ProjThresh(in_files=ldir, threshold=3)
    >>> pThresh.cmdline
    'proj_thresh seeds_to_M1.nii seeds_to_M2.nii 3'

    """

    _cmd = "proj_thresh"
    input_spec = ProjThreshInputSpec
    output_spec = ProjThreshOuputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_files"] = []
        for name in self.inputs.in_files:
            cwd, base_name = os.path.split(name)
            outputs["out_files"].append(
                self._gen_fname(
                    base_name,
                    cwd=cwd,
                    suffix="_proj_seg_thr_{}".format(self.inputs.threshold),
                )
            )
        return outputs


class FindTheBiggestInputSpec(FSLCommandInputSpec):
    in_files = traits.List(
        File(exists=True),
        argstr="%s",
        desc=("a list of input volumes or a " "singleMatrixFile"),
        position=0,
        mandatory=True,
    )
    out_file = File(
        argstr="%s",
        desc="file with the resulting segmentation",
        position=2,
        genfile=True,
        hash_files=False,
    )


class FindTheBiggestOutputSpec(TraitedSpec):
    out_file = File(
        exists=True, argstr="%s", desc="output file indexed in order of input files"
    )


class FindTheBiggest(FSLCommand):
    """
    Use FSL find_the_biggest for performing hard segmentation on
    the outputs of connectivity-based thresholding in probtrack.
    For complete details, see the `FDT
    Documentation. <http://www.fmrib.ox.ac.uk/fsl/fdt/fdt_biggest.html>`_

    Example
    -------

    >>> from nipype.interfaces import fsl
    >>> ldir = ['seeds_to_M1.nii', 'seeds_to_M2.nii']
    >>> fBig = fsl.FindTheBiggest(in_files=ldir, out_file='biggestSegmentation')
    >>> fBig.cmdline
    'find_the_biggest seeds_to_M1.nii seeds_to_M2.nii biggestSegmentation'

    """

    _cmd = "find_the_biggest"
    input_spec = FindTheBiggestInputSpec
    output_spec = FindTheBiggestOutputSpec

    def _run_interface(self, runtime):
        if not isdefined(self.inputs.out_file):
            self.inputs.out_file = self._gen_fname("biggestSegmentation", suffix="")
        return super(FindTheBiggest, self)._run_interface(runtime)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_file
        if not isdefined(outputs["out_file"]):
            outputs["out_file"] = self._gen_fname("biggestSegmentation", suffix="")
        outputs["out_file"] = os.path.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._list_outputs()[name]
        else:
            return None


class TractSkeletonInputSpec(FSLCommandInputSpec):

    in_file = File(
        exists=True,
        mandatory=True,
        argstr="-i %s",
        desc="input image (typcially mean FA volume)",
    )
    _proj_inputs = ["threshold", "distance_map", "data_file"]
    project_data = traits.Bool(
        argstr="-p %.3f %s %s %s %s",
        requires=_proj_inputs,
        desc="project data onto skeleton",
    )
    threshold = traits.Float(desc="skeleton threshold value")
    distance_map = File(exists=True, desc="distance map image")
    search_mask_file = File(
        exists=True,
        xor=["use_cingulum_mask"],
        desc="mask in which to use alternate search rule",
    )
    use_cingulum_mask = traits.Bool(
        True,
        usedefault=True,
        xor=["search_mask_file"],
        desc=("perform alternate search using " "built-in cingulum mask"),
    )
    data_file = File(exists=True, desc="4D data to project onto skeleton (usually FA)")
    alt_data_file = File(
        exists=True, argstr="-a %s", desc="4D non-FA data to project onto skeleton"
    )
    alt_skeleton = File(exists=True, argstr="-s %s", desc="alternate skeleton to use")
    projected_data = File(desc="input data projected onto skeleton")
    skeleton_file = traits.Either(
        traits.Bool, File, argstr="-o %s", desc="write out skeleton image"
    )


class TractSkeletonOutputSpec(TraitedSpec):

    projected_data = File(desc="input data projected onto skeleton")
    skeleton_file = File(desc="tract skeleton image")


class TractSkeleton(FSLCommand):
    """Use FSL's tbss_skeleton to skeletonise an FA image or project arbitrary
    values onto a skeleton.

    There are two ways to use this interface.  To create a skeleton from an FA
    image, just supply the ``in_file`` and set ``skeleton_file`` to True (or
    specify a skeleton filename. To project values onto a skeleton, you must
    set ``project_data`` to True, and then also supply values for
    ``threshold``, ``distance_map``, and ``data_file``. The
    ``search_mask_file`` and ``use_cingulum_mask`` inputs are also used in data
    projection, but ``use_cingulum_mask`` is set to True by default.  This mask
    controls where the projection algorithm searches within a circular space
    around a tract, rather than in a single perpindicular direction.

    Example
    -------

    >>> import nipype.interfaces.fsl as fsl
    >>> skeletor = fsl.TractSkeleton()
    >>> skeletor.inputs.in_file = "all_FA.nii.gz"
    >>> skeletor.inputs.skeleton_file = True
    >>> skeletor.run() # doctest: +SKIP

    """

    _cmd = "tbss_skeleton"
    input_spec = TractSkeletonInputSpec
    output_spec = TractSkeletonOutputSpec

    def _format_arg(self, name, spec, value):
        if name == "project_data":
            if isdefined(value) and value:
                _si = self.inputs
                if isdefined(_si.use_cingulum_mask) and _si.use_cingulum_mask:
                    mask_file = Info.standard_image("LowerCingulum_1mm.nii.gz")
                else:
                    mask_file = _si.search_mask_file
                if not isdefined(_si.projected_data):
                    proj_file = self._list_outputs()["projected_data"]
                else:
                    proj_file = _si.projected_data
                return spec.argstr % (
                    _si.threshold,
                    _si.distance_map,
                    mask_file,
                    _si.data_file,
                    proj_file,
                )
        elif name == "skeleton_file":
            if isinstance(value, bool):
                return spec.argstr % self._list_outputs()["skeleton_file"]
            else:
                return spec.argstr % value
        return super(TractSkeleton, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        _si = self.inputs
        if isdefined(_si.project_data) and _si.project_data:
            proj_data = _si.projected_data
            outputs["projected_data"] = proj_data
            if not isdefined(proj_data):
                stem = _si.data_file
                if isdefined(_si.alt_data_file):
                    stem = _si.alt_data_file
                outputs["projected_data"] = fname_presuffix(
                    stem, suffix="_skeletonised", newpath=os.getcwd(), use_ext=True
                )
        if isdefined(_si.skeleton_file) and _si.skeleton_file:
            outputs["skeleton_file"] = _si.skeleton_file
            if isinstance(_si.skeleton_file, bool):
                outputs["skeleton_file"] = fname_presuffix(
                    _si.in_file, suffix="_skeleton", newpath=os.getcwd(), use_ext=True
                )
        return outputs


class DistanceMapInputSpec(FSLCommandInputSpec):

    in_file = File(
        exists=True,
        mandatory=True,
        argstr="--in=%s",
        desc="image to calculate distance values for",
    )
    mask_file = File(
        exists=True, argstr="--mask=%s", desc="binary mask to contrain calculations"
    )
    invert_input = traits.Bool(argstr="--invert", desc="invert input image")
    local_max_file = traits.Either(
        traits.Bool,
        File,
        argstr="--localmax=%s",
        desc="write an image of the local maxima",
        hash_files=False,
    )
    distance_map = File(
        genfile=True, argstr="--out=%s", desc="distance map to write", hash_files=False
    )


class DistanceMapOutputSpec(TraitedSpec):

    distance_map = File(exists=True, desc="value is distance to nearest nonzero voxels")
    local_max_file = File(desc="image of local maxima")


class DistanceMap(FSLCommand):
    """Use FSL's distancemap to generate a map of the distance to the nearest
    nonzero voxel.

    Example
    -------

    >>> import nipype.interfaces.fsl as fsl
    >>> mapper = fsl.DistanceMap()
    >>> mapper.inputs.in_file = "skeleton_mask.nii.gz"
    >>> mapper.run() # doctest: +SKIP

    """

    _cmd = "distancemap"
    input_spec = DistanceMapInputSpec
    output_spec = DistanceMapOutputSpec

    def _format_arg(self, name, spec, value):
        if name == "local_max_file":
            if isinstance(value, bool):
                return spec.argstr % self._list_outputs()["local_max_file"]
        return super(DistanceMap, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        _si = self.inputs
        outputs["distance_map"] = _si.distance_map
        if not isdefined(_si.distance_map):
            outputs["distance_map"] = fname_presuffix(
                _si.in_file, suffix="_dstmap", use_ext=True, newpath=os.getcwd()
            )
        outputs["distance_map"] = os.path.abspath(outputs["distance_map"])
        if isdefined(_si.local_max_file):
            outputs["local_max_file"] = _si.local_max_file
            if isinstance(_si.local_max_file, bool):
                outputs["local_max_file"] = fname_presuffix(
                    _si.in_file, suffix="_lclmax", use_ext=True, newpath=os.getcwd()
                )
            outputs["local_max_file"] = os.path.abspath(outputs["local_max_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "distance_map":
            return self._list_outputs()["distance_map"]
        return None


class MakeDyadicVectorsInputSpec(FSLCommandInputSpec):
    theta_vol = File(exists=True, mandatory=True, position=0, argstr="%s")
    phi_vol = File(exists=True, mandatory=True, position=1, argstr="%s")
    mask = File(exists=True, position=2, argstr="%s")
    output = File("dyads", position=3, usedefault=True, argstr="%s", hash_files=False)
    perc = traits.Float(
        desc=(
            "the {perc}% angle of the output cone of "
            "uncertainty (output will be in degrees)"
        ),
        position=4,
        argstr="%f",
    )


class MakeDyadicVectorsOutputSpec(TraitedSpec):
    dyads = File(exists=True)
    dispersion = File(exists=True)


class MakeDyadicVectors(FSLCommand):
    """Create vector volume representing mean principal diffusion direction
    and its uncertainty (dispersion)"""

    _cmd = "make_dyadic_vectors"
    input_spec = MakeDyadicVectorsInputSpec
    output_spec = MakeDyadicVectorsOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["dyads"] = self._gen_fname(self.inputs.output)
        outputs["dispersion"] = self._gen_fname(
            self.inputs.output, suffix="_dispersion"
        )

        return outputs



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os

import pytest
from nipype.testing.fixtures import create_files_in_directory

import nipype.interfaces.spm as spm
from nipype.interfaces.spm import no_spm
import nipype.interfaces.matlab as mlab

mlab.MatlabCommand.set_default_matlab_cmd(os.getenv("MATLABCMD", "matlab"))


def test_slicetiming():
    assert spm.SliceTiming._jobtype == "temporal"
    assert spm.SliceTiming._jobname == "st"


def test_slicetiming_list_outputs(create_files_in_directory):
    filelist, outdir = create_files_in_directory
    st = spm.SliceTiming(in_files=filelist[0])
    assert st._list_outputs()["timecorrected_files"][0][0] == "a"


def test_realign():
    assert spm.Realign._jobtype == "spatial"
    assert spm.Realign._jobname == "realign"
    assert spm.Realign().inputs.jobtype == "estwrite"


def test_realign_list_outputs(create_files_in_directory):
    filelist, outdir = create_files_in_directory
    rlgn = spm.Realign(in_files=filelist[0])
    assert rlgn._list_outputs()["realignment_parameters"][0].startswith("rp_")
    assert rlgn._list_outputs()["realigned_files"][0].startswith("r")
    assert rlgn._list_outputs()["mean_image"].startswith("mean")


def test_coregister():
    assert spm.Coregister._jobtype == "spatial"
    assert spm.Coregister._jobname == "coreg"
    assert spm.Coregister().inputs.jobtype == "estwrite"


def test_coregister_list_outputs(create_files_in_directory):
    filelist, outdir = create_files_in_directory
    coreg = spm.Coregister(source=filelist[0])
    assert coreg._list_outputs()["coregistered_source"][0].startswith("r")
    coreg = spm.Coregister(source=filelist[0], apply_to_files=filelist[1])
    assert coreg._list_outputs()["coregistered_files"][0].startswith("r")


def test_normalize():
    assert spm.Normalize._jobtype == "spatial"
    assert spm.Normalize._jobname == "normalise"
    assert spm.Normalize().inputs.jobtype == "estwrite"


def test_normalize_list_outputs(create_files_in_directory):
    filelist, outdir = create_files_in_directory
    norm = spm.Normalize(source=filelist[0])
    assert norm._list_outputs()["normalized_source"][0].startswith("w")
    norm = spm.Normalize(source=filelist[0], apply_to_files=filelist[1])
    assert norm._list_outputs()["normalized_files"][0].startswith("w")


def test_normalize12():
    assert spm.Normalize12._jobtype == "spatial"
    assert spm.Normalize12._jobname == "normalise"
    assert spm.Normalize12().inputs.jobtype == "estwrite"


def test_normalize12_list_outputs(create_files_in_directory):
    filelist, outdir = create_files_in_directory
    norm12 = spm.Normalize12(image_to_align=filelist[0])
    assert norm12._list_outputs()["normalized_image"][0].startswith("w")
    norm12 = spm.Normalize12(image_to_align=filelist[0], apply_to_files=filelist[1])
    assert norm12._list_outputs()["normalized_files"][0].startswith("w")


@pytest.mark.skipif(no_spm(), reason="spm is not installed")
def test_segment():
    if spm.Info.name() == "SPM12":
        assert spm.Segment()._jobtype == "tools"
        assert spm.Segment()._jobname == "oldseg"
    else:
        assert spm.Segment()._jobtype == "spatial"
        assert spm.Segment()._jobname == "preproc"


@pytest.mark.skipif(no_spm(), reason="spm is not installed")
def test_newsegment():
    if spm.Info.name() == "SPM12":
        assert spm.NewSegment()._jobtype == "spatial"
        assert spm.NewSegment()._jobname == "preproc"
    else:
        assert spm.NewSegment()._jobtype == "tools"
        assert spm.NewSegment()._jobname == "preproc8"


def test_smooth():
    assert spm.Smooth._jobtype == "spatial"
    assert spm.Smooth._jobname == "smooth"


def test_dartel():
    assert spm.DARTEL._jobtype == "tools"
    assert spm.DARTEL._jobname == "dartel"


def test_dartelnorm2mni():
    assert spm.DARTELNorm2MNI._jobtype == "tools"
    assert spm.DARTELNorm2MNI._jobname == "dartel"



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The minc module provides classes for interfacing with the `MINC
<http://www.bic.mni.mcgill.ca/ServicesSoftware/MINC>`_ command line tools.
This module was written to work with MINC version 2.2.00.

Author: Carlo Hamalainen <carlo@carlo-hamalainen.net>
        http://carlo-hamalainen.net
"""
import os
import os.path
import warnings

from ..base import CommandLine

warnings.filterwarnings("always", category=UserWarning)


def check_minc():
    """Returns True if and only if MINC is installed.'
    """

    return Info.version() is not None


def no_minc():
    """Returns True if and only if MINC is *not* installed.
    """
    return not check_minc()


class Info(object):
    """Handle MINC version information.

    version refers to the version of MINC on the system
    """

    @staticmethod
    def version():
        """Check for minc version on the system

        Parameters
        ----------
        None

        Returns
        -------
        version : dict
           Version number as dict or None if MINC not found

        """
        try:
            clout = CommandLine(
                command="mincinfo", args="-version", terminal_output="allatonce"
            ).run()
        except IOError:
            return None

        out = clout.runtime.stdout

        def read_program_version(s):
            if "program" in s:
                return s.split(":")[1].strip()
            return None

        def read_libminc_version(s):
            if "libminc" in s:
                return s.split(":")[1].strip()
            return None

        def read_netcdf_version(s):
            if "netcdf" in s:
                return " ".join(s.split(":")[1:]).strip()
            return None

        def read_hdf5_version(s):
            if "HDF5" in s:
                return s.split(":")[1].strip()
            return None

        versions = {
            "minc": None,
            "libminc": None,
            "netcdf": None,
            "hdf5": None,
        }

        for l in out.split("\n"):
            for (name, f) in [
                ("minc", read_program_version),
                ("libminc", read_libminc_version),
                ("netcdf", read_netcdf_version),
                ("hdf5", read_hdf5_version),
            ]:
                if f(l) is not None:
                    versions[name] = f(l)

        return versions


def aggregate_filename(files, new_suffix):
    """
    Try to work out a sensible name given a set of files that have
    been combined in some way (e.g. averaged). If we can't work out a
    sensible prefix, we use the first filename in the list.

    Examples
    --------

    >>> from nipype.interfaces.minc.base import aggregate_filename
    >>> f = aggregate_filename(['/tmp/foo1.mnc', '/tmp/foo2.mnc', '/tmp/foo3.mnc'], 'averaged')
    >>> os.path.split(f)[1] # This has a full path, so just check the filename.
    'foo_averaged.mnc'

    >>> f = aggregate_filename(['/tmp/foo1.mnc', '/tmp/blah1.mnc'], 'averaged')
    >>> os.path.split(f)[1] # This has a full path, so just check the filename.
    'foo1_averaged.mnc'

    """

    path = os.path.split(files[0])[0]
    names = [os.path.splitext(os.path.split(x)[1])[0] for x in files]
    common_prefix = os.path.commonprefix(names)

    path = os.getcwd()

    if common_prefix == "":
        return os.path.abspath(
            os.path.join(
                path, os.path.splitext(files[0])[0] + "_" + new_suffix + ".mnc"
            )
        )
    else:
        return os.path.abspath(
            os.path.join(path, common_prefix + "_" + new_suffix + ".mnc")
        )



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import AlignEpiAnatPy


def test_AlignEpiAnatPy_inputs():
    input_map = dict(
        anat=dict(
            argstr="-anat %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        anat2epi=dict(
            argstr="-anat2epi",
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        epi2anat=dict(
            argstr="-epi2anat",
        ),
        epi_base=dict(
            argstr="-epi_base %s",
            mandatory=True,
        ),
        epi_strip=dict(
            argstr="-epi_strip %s",
        ),
        in_file=dict(
            argstr="-epi %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        outputtype=dict(),
        py27_path=dict(
            usedefault=True,
        ),
        save_skullstrip=dict(
            argstr="-save_skullstrip",
        ),
        suffix=dict(
            argstr="-suffix %s",
            usedefault=True,
        ),
        tshift=dict(
            argstr="-tshift %s",
            usedefault=True,
        ),
        volreg=dict(
            argstr="-volreg %s",
            usedefault=True,
        ),
    )
    inputs = AlignEpiAnatPy.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_AlignEpiAnatPy_outputs():
    output_map = dict(
        anat_al_mat=dict(
            extensions=None,
        ),
        anat_al_orig=dict(
            extensions=None,
        ),
        epi_al_mat=dict(
            extensions=None,
        ),
        epi_al_orig=dict(
            extensions=None,
        ),
        epi_al_tlrc_mat=dict(
            extensions=None,
        ),
        epi_reg_al_mat=dict(
            extensions=None,
        ),
        epi_tlrc_al=dict(
            extensions=None,
        ),
        epi_vr_al_mat=dict(
            extensions=None,
        ),
        epi_vr_motion=dict(
            extensions=None,
        ),
        skullstrip=dict(
            extensions=None,
        ),
    )
    outputs = AlignEpiAnatPy.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from .denoising import UnbiasedNonLocalMeans
from .featuredetection import (
    GenerateSummedGradientImage,
    CannySegmentationLevelSetImageFilter,
    DilateImage,
    TextureFromNoiseImageFilter,
    FlippedDifference,
    ErodeImage,
    GenerateBrainClippedImage,
    NeighborhoodMedian,
    GenerateTestImage,
    NeighborhoodMean,
    HammerAttributeCreator,
    TextureMeasureFilter,
    DilateMask,
    DumpBinaryTrainingVectors,
    DistanceMaps,
    STAPLEAnalysis,
    GradientAnisotropicDiffusionImageFilter,
    CannyEdge,
)



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)
import os


class RobustStatisticsSegmenterInputSpec(CommandLineInputSpec):
    expectedVolume = traits.Float(
        desc="The approximate volume of the object, in mL.",
        argstr="--expectedVolume %f",
    )
    intensityHomogeneity = traits.Float(
        desc="What is the homogeneity of intensity within the object? Given constant intensity at 1.0 score and extreme fluctuating intensity at 0.",
        argstr="--intensityHomogeneity %f",
    )
    curvatureWeight = traits.Float(
        desc="Given sphere 1.0 score and extreme rough bounday/surface 0 score, what is the expected smoothness of the object?",
        argstr="--curvatureWeight %f",
    )
    labelValue = traits.Int(
        desc="Label value of the output image", argstr="--labelValue %d"
    )
    maxRunningTime = traits.Float(
        desc="The program will stop if this time is reached.",
        argstr="--maxRunningTime %f",
    )
    originalImageFileName = File(
        position=-3, desc="Original image to be segmented", exists=True, argstr="%s"
    )
    labelImageFileName = File(
        position=-2, desc="Label image for initialization", exists=True, argstr="%s"
    )
    segmentedImageFileName = traits.Either(
        traits.Bool,
        File(),
        position=-1,
        hash_files=False,
        desc="Segmented image",
        argstr="%s",
    )


class RobustStatisticsSegmenterOutputSpec(TraitedSpec):
    segmentedImageFileName = File(position=-1, desc="Segmented image", exists=True)


class RobustStatisticsSegmenter(SEMLikeCommandLine):
    """title: Robust Statistics Segmenter

category: Segmentation.Specialized

description: Active contour segmentation using robust statistic.

version: 1.0

documentation-url: http://wiki.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/RobustStatisticsSegmenter

contributor: Yi Gao (gatech), Allen Tannenbaum (gatech), Ron Kikinis (SPL, BWH)

acknowledgements: This work is part of the National Alliance for Medical Image Computing (NAMIC), funded by the National Institutes of Health

"""

    input_spec = RobustStatisticsSegmenterInputSpec
    output_spec = RobustStatisticsSegmenterOutputSpec
    _cmd = "RobustStatisticsSegmenter "
    _outputs_filenames = {"segmentedImageFileName": "segmentedImageFileName.nii"}


class EMSegmentCommandLineInputSpec(CommandLineInputSpec):
    mrmlSceneFileName = File(
        desc="Active MRML scene that contains EMSegment algorithm parameters.",
        exists=True,
        argstr="--mrmlSceneFileName %s",
    )
    resultVolumeFileName = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The file name that the segmentation result volume will be written to.",
        argstr="--resultVolumeFileName %s",
    )
    targetVolumeFileNames = InputMultiPath(
        File(exists=True),
        desc="File names of target volumes (to be segmented).  The number of target images must be equal to the number of target images specified in the parameter set, and these images must be spatially aligned.",
        argstr="--targetVolumeFileNames %s...",
    )
    intermediateResultsDirectory = Directory(
        desc="Directory where EMSegmenter will write intermediate data (e.g., aligned atlas data).",
        exists=True,
        argstr="--intermediateResultsDirectory %s",
    )
    parametersMRMLNodeName = traits.Str(
        desc="The name of the EMSegment parameters node within the active MRML scene.  Leave blank for default.",
        argstr="--parametersMRMLNodeName %s",
    )
    disableMultithreading = traits.Int(
        desc="Disable multithreading for the EMSegmenter algorithm only! Preprocessing might still run in multi-threaded mode. -1: Do not overwrite default value. 0: Disable. 1: Enable.",
        argstr="--disableMultithreading %d",
    )
    dontUpdateIntermediateData = traits.Int(
        desc="Disable update of intermediate results.  -1: Do not overwrite default value. 0: Disable. 1: Enable.",
        argstr="--dontUpdateIntermediateData %d",
    )
    verbose = traits.Bool(desc="Enable verbose output.", argstr="--verbose ")
    loadTargetCentered = traits.Bool(
        desc="Read target files centered.", argstr="--loadTargetCentered "
    )
    loadAtlasNonCentered = traits.Bool(
        desc="Read atlas files non-centered.", argstr="--loadAtlasNonCentered "
    )
    taskPreProcessingSetting = traits.Str(
        desc="Specifies the different task parameter. Leave blank for default.",
        argstr="--taskPreProcessingSetting %s",
    )
    keepTempFiles = traits.Bool(
        desc="If flag is set then at the end of command the temporary files are not removed",
        argstr="--keepTempFiles ",
    )
    resultStandardVolumeFileName = File(
        desc="Used for testing.  Compare segmentation results to this image and return EXIT_FAILURE if they do not match.",
        exists=True,
        argstr="--resultStandardVolumeFileName %s",
    )
    dontWriteResults = traits.Bool(
        desc="Used for testing.  Don't actually write the resulting labelmap to disk.",
        argstr="--dontWriteResults ",
    )
    generateEmptyMRMLSceneAndQuit = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Used for testing.  Only write a scene with default mrml parameters.",
        argstr="--generateEmptyMRMLSceneAndQuit %s",
    )
    resultMRMLSceneFileName = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Write out the MRML scene after command line substitutions have been made.",
        argstr="--resultMRMLSceneFileName %s",
    )
    disableCompression = traits.Bool(
        desc="Don't use compression when writing result image to disk.",
        argstr="--disableCompression ",
    )
    atlasVolumeFileNames = InputMultiPath(
        File(exists=True),
        desc="Use an alternative atlas to the one that is specified by the mrml file - note the order matters ! ",
        argstr="--atlasVolumeFileNames %s...",
    )
    registrationPackage = traits.Str(
        desc="specify the registration package for preprocessing (CMTK or BRAINS or PLASTIMATCH or DEMONS)",
        argstr="--registrationPackage %s",
    )
    registrationAffineType = traits.Int(
        desc="specify the accuracy of the affine registration. -2: Do not overwrite default, -1: Test, 0: Disable, 1: Fast, 2: Accurate",
        argstr="--registrationAffineType %d",
    )
    registrationDeformableType = traits.Int(
        desc="specify the accuracy of the deformable registration. -2: Do not overwrite default, -1: Test, 0: Disable, 1: Fast, 2: Accurate",
        argstr="--registrationDeformableType %d",
    )


class EMSegmentCommandLineOutputSpec(TraitedSpec):
    resultVolumeFileName = File(
        desc="The file name that the segmentation result volume will be written to.",
        exists=True,
    )
    generateEmptyMRMLSceneAndQuit = File(
        desc="Used for testing.  Only write a scene with default mrml parameters.",
        exists=True,
    )
    resultMRMLSceneFileName = File(
        desc="Write out the MRML scene after command line substitutions have been made.",
        exists=True,
    )


class EMSegmentCommandLine(SEMLikeCommandLine):
    """title:
  EMSegment Command-line


category:
  Segmentation.Specialized


description:
  This module is used to simplify the process of segmenting large collections of images by providing a command line interface to the EMSegment algorithm for script and batch processing.


documentation-url: http://www.slicer.org/slicerWiki/index.php/Documentation/4.0/EMSegment_Command-line

contributor: Sebastien Barre, Brad Davis, Kilian Pohl, Polina Golland, Yumin Yuan, Daniel Haehn

acknowledgements: Many people and organizations have contributed to the funding, design, and development of the EMSegment algorithm and its various implementations.


"""

    input_spec = EMSegmentCommandLineInputSpec
    output_spec = EMSegmentCommandLineOutputSpec
    _cmd = "EMSegmentCommandLine "
    _outputs_filenames = {
        "generateEmptyMRMLSceneAndQuit": "generateEmptyMRMLSceneAndQuit",
        "resultMRMLSceneFileName": "resultMRMLSceneFileName",
        "resultVolumeFileName": "resultVolumeFileName.mhd",
    }


class BRAINSROIAutoInputSpec(CommandLineInputSpec):
    inputVolume = File(
        desc="The input image for finding the largest region filled mask.",
        exists=True,
        argstr="--inputVolume %s",
    )
    outputROIMaskVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The ROI automatically found from the input image.",
        argstr="--outputROIMaskVolume %s",
    )
    outputClippedVolumeROI = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The inputVolume clipped to the region of the brain mask.",
        argstr="--outputClippedVolumeROI %s",
    )
    otsuPercentileThreshold = traits.Float(
        desc="Parameter to the Otsu threshold algorithm.",
        argstr="--otsuPercentileThreshold %f",
    )
    thresholdCorrectionFactor = traits.Float(
        desc="A factor to scale the Otsu algorithm's result threshold, in case clipping mangles the image.",
        argstr="--thresholdCorrectionFactor %f",
    )
    closingSize = traits.Float(
        desc="The Closing Size (in millimeters) for largest connected filled mask.  This value is divided by image spacing and rounded to the next largest voxel number.",
        argstr="--closingSize %f",
    )
    ROIAutoDilateSize = traits.Float(
        desc="This flag is only relavent when using ROIAUTO mode for initializing masks.  It defines the final dilation size to capture a bit of background outside the tissue region.  At setting of 10mm has been shown to help regularize a BSpline registration type so that there is some background constraints to match the edges of the head better.",
        argstr="--ROIAutoDilateSize %f",
    )
    outputVolumePixelType = traits.Enum(
        "float",
        "short",
        "ushort",
        "int",
        "uint",
        "uchar",
        desc="The output image Pixel Type is the scalar datatype for representation of the Output Volume.",
        argstr="--outputVolumePixelType %s",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSROIAutoOutputSpec(TraitedSpec):
    outputROIMaskVolume = File(
        desc="The ROI automatically found from the input image.", exists=True
    )
    outputClippedVolumeROI = File(
        desc="The inputVolume clipped to the region of the brain mask.", exists=True
    )


class BRAINSROIAuto(SEMLikeCommandLine):
    """title: Foreground masking (BRAINS)

category: Segmentation.Specialized

description: This tool uses a combination of otsu thresholding and a closing operations to identify the most prominant foreground region in an image.


version: 2.4.1

license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

contributor: Hans J. Johnson, hans-johnson -at- uiowa.edu, http://wwww.psychiatry.uiowa.edu

acknowledgements: Hans Johnson(1,3,4); Kent Williams(1); Gregory Harris(1), Vincent Magnotta(1,2,3);  Andriy Fedorov(5), fedorov -at- bwh.harvard.edu (Slicer integration); (1=University of Iowa Department of Psychiatry, 2=University of Iowa Department of Radiology, 3=University of Iowa Department of Biomedical Engineering, 4=University of Iowa Department of Electrical and Computer Engineering, 5=Surgical Planning Lab, Harvard)

"""

    input_spec = BRAINSROIAutoInputSpec
    output_spec = BRAINSROIAutoOutputSpec
    _cmd = "BRAINSROIAuto "
    _outputs_filenames = {
        "outputROIMaskVolume": "outputROIMaskVolume.nii",
        "outputClippedVolumeROI": "outputClippedVolumeROI.nii",
    }



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-

import os.path as op

from ..base import (
    CommandLineInputSpec,
    CommandLine,
    traits,
    TraitedSpec,
    File,
    InputMultiPath,
    isdefined,
)
from .base import MRTrix3BaseInputSpec, MRTrix3Base


class BrainMaskInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="input diffusion weighted images",
    )
    out_file = File(
        "brainmask.mif",
        argstr="%s",
        mandatory=True,
        position=-1,
        usedefault=True,
        desc="output brain mask",
    )


class BrainMaskOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output response file")


class BrainMask(CommandLine):
    """
    Convert a mesh surface to a partial volume estimation image


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> bmsk = mrt.BrainMask()
    >>> bmsk.inputs.in_file = 'dwi.mif'
    >>> bmsk.cmdline                               # doctest: +ELLIPSIS
    'dwi2mask dwi.mif brainmask.mif'
    >>> bmsk.run()                                 # doctest: +SKIP
    """

    _cmd = "dwi2mask"
    input_spec = BrainMaskInputSpec
    output_spec = BrainMaskOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class Mesh2PVEInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-3, desc="input mesh"
    )
    reference = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="input reference image",
    )
    in_first = File(
        exists=True,
        argstr="-first %s",
        desc="indicates that the mesh file is provided by FSL FIRST",
    )

    out_file = File(
        "mesh2volume.nii.gz",
        argstr="%s",
        mandatory=True,
        position=-1,
        usedefault=True,
        desc="output file containing SH coefficients",
    )


class Mesh2PVEOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output response file")


class Mesh2PVE(CommandLine):
    """
    Convert a mesh surface to a partial volume estimation image


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> m2p = mrt.Mesh2PVE()
    >>> m2p.inputs.in_file = 'surf1.vtk'
    >>> m2p.inputs.reference = 'dwi.mif'
    >>> m2p.inputs.in_first = 'T1.nii.gz'
    >>> m2p.cmdline                               # doctest: +ELLIPSIS
    'mesh2pve -first T1.nii.gz surf1.vtk dwi.mif mesh2volume.nii.gz'
    >>> m2p.run()                                 # doctest: +SKIP
    """

    _cmd = "mesh2pve"
    input_spec = Mesh2PVEInputSpec
    output_spec = Mesh2PVEOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class Generate5ttInputSpec(MRTrix3BaseInputSpec):
    algorithm = traits.Enum(
        "fsl",
        "gif",
        "freesurfer",
        argstr="%s",
        position=-3,
        mandatory=True,
        desc="tissue segmentation algorithm",
    )
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-2, desc="input image"
    )
    out_file = File(argstr="%s", mandatory=True, position=-1, desc="output image")


class Generate5ttOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="output image")


class Generate5tt(MRTrix3Base):
    """
    Generate a 5TT image suitable for ACT using the selected algorithm


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> gen5tt = mrt.Generate5tt()
    >>> gen5tt.inputs.in_file = 'T1.nii.gz'
    >>> gen5tt.inputs.algorithm = 'fsl'
    >>> gen5tt.inputs.out_file = '5tt.mif'
    >>> gen5tt.cmdline                             # doctest: +ELLIPSIS
    '5ttgen fsl T1.nii.gz 5tt.mif'
    >>> gen5tt.run()                               # doctest: +SKIP
    """

    _cmd = "5ttgen"
    input_spec = Generate5ttInputSpec
    output_spec = Generate5ttOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class TensorMetricsInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-1, desc="input DTI image"
    )

    out_fa = File(argstr="-fa %s", desc="output FA file")
    out_adc = File(argstr="-adc %s", desc="output ADC file")
    out_evec = File(argstr="-vector %s", desc="output selected eigenvector(s) file")
    out_eval = File(argstr="-value %s", desc="output selected eigenvalue(s) file")
    component = traits.List(
        [1],
        usedefault=True,
        argstr="-num %s",
        sep=",",
        desc=(
            "specify the desired eigenvalue/eigenvector(s). Note that "
            "several eigenvalues can be specified as a number sequence"
        ),
    )
    in_mask = File(
        exists=True,
        argstr="-mask %s",
        desc=(
            "only perform computation within the specified binary" " brain mask image"
        ),
    )
    modulate = traits.Enum(
        "FA",
        "none",
        "eval",
        argstr="-modulate %s",
        desc=("how to modulate the magnitude of the" " eigenvectors"),
    )


class TensorMetricsOutputSpec(TraitedSpec):
    out_fa = File(desc="output FA file")
    out_adc = File(desc="output ADC file")
    out_evec = File(desc="output selected eigenvector(s) file")
    out_eval = File(desc="output selected eigenvalue(s) file")


class TensorMetrics(CommandLine):
    """
    Compute metrics from tensors


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> comp = mrt.TensorMetrics()
    >>> comp.inputs.in_file = 'dti.mif'
    >>> comp.inputs.out_fa = 'fa.mif'
    >>> comp.cmdline                               # doctest: +ELLIPSIS
    'tensor2metric -num 1 -fa fa.mif dti.mif'
    >>> comp.run()                                 # doctest: +SKIP
    """

    _cmd = "tensor2metric"
    input_spec = TensorMetricsInputSpec
    output_spec = TensorMetricsOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()

        for k in list(outputs.keys()):
            if isdefined(getattr(self.inputs, k)):
                outputs[k] = op.abspath(getattr(self.inputs, k))

        return outputs


class ComputeTDIInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-2, desc="input tractography"
    )
    out_file = File(
        "tdi.mif", argstr="%s", usedefault=True, position=-1, desc="output TDI file"
    )
    reference = File(
        exists=True,
        argstr="-template %s",
        desc="a reference" "image to be used as template",
    )
    vox_size = traits.List(
        traits.Int, argstr="-vox %s", sep=",", desc="voxel dimensions"
    )
    data_type = traits.Enum(
        "float",
        "unsigned int",
        argstr="-datatype %s",
        desc="specify output image data type",
    )
    use_dec = traits.Bool(argstr="-dec", desc="perform mapping in DEC space")
    dixel = File(
        argstr="-dixel %s",
        desc="map streamlines to"
        "dixels within each voxel. Directions are stored as"
        "azimuth elevation pairs.",
    )
    max_tod = traits.Int(
        argstr="-tod %d",
        desc="generate a Track Orientation " "Distribution (TOD) in each voxel.",
    )

    contrast = traits.Enum(
        "tdi",
        "length",
        "invlength",
        "scalar_map",
        "scalar_map_conut",
        "fod_amp",
        "curvature",
        argstr="-constrast %s",
        desc="define the desired " "form of contrast for the output image",
    )
    in_map = File(
        exists=True,
        argstr="-image %s",
        desc="provide the"
        "scalar image map for generating images with "
        "'scalar_map' contrasts, or the SHs image for fod_amp",
    )

    stat_vox = traits.Enum(
        "sum",
        "min",
        "mean",
        "max",
        argstr="-stat_vox %s",
        desc="define the statistic for choosing the final"
        "voxel intesities for a given contrast",
    )
    stat_tck = traits.Enum(
        "mean",
        "sum",
        "min",
        "max",
        "median",
        "mean_nonzero",
        "gaussian",
        "ends_min",
        "ends_mean",
        "ends_max",
        "ends_prod",
        argstr="-stat_tck %s",
        desc="define the statistic for choosing "
        "the contribution to be made by each streamline as a function of"
        " the samples taken along their lengths.",
    )

    fwhm_tck = traits.Float(
        argstr="-fwhm_tck %f",
        desc="define the statistic for choosing the"
        " contribution to be made by each streamline as a function of the "
        "samples taken along their lengths",
    )

    map_zero = traits.Bool(
        argstr="-map_zero",
        desc="if a streamline has zero contribution based "
        "on the contrast & statistic, typically it is not mapped; use this "
        "option to still contribute to the map even if this is the case "
        "(these non-contributing voxels can then influence the mean value in "
        "each voxel of the map)",
    )

    upsample = traits.Int(
        argstr="-upsample %d",
        desc="upsample the tracks by"
        " some ratio using Hermite interpolation before "
        "mappping",
    )

    precise = traits.Bool(
        argstr="-precise",
        desc="use a more precise streamline mapping "
        "strategy, that accurately quantifies the length through each voxel "
        "(these lengths are then taken into account during TWI calculation)",
    )
    ends_only = traits.Bool(
        argstr="-ends_only", desc="only map the streamline" " endpoints to the image"
    )

    tck_weights = File(
        exists=True,
        argstr="-tck_weights_in %s",
        desc="specify" " a text scalar file containing the streamline weights",
    )
    nthreads = traits.Int(
        argstr="-nthreads %d",
        desc="number of threads. if zero, the number" " of available cpus will be used",
        nohash=True,
    )


class ComputeTDIOutputSpec(TraitedSpec):
    out_file = File(desc="output TDI file")


class ComputeTDI(MRTrix3Base):
    """
    Use track data as a form of contrast for producing a high-resolution
    image.

    .. admonition:: References

      * For TDI or DEC TDI: Calamante, F.; Tournier, J.-D.; Jackson, G. D. &
        Connelly, A. Track-density imaging (TDI): Super-resolution white
        matter imaging using whole-brain track-density mapping. NeuroImage,
        2010, 53, 1233-1243

      * If using -contrast length and -stat_vox mean: Pannek, K.; Mathias,
        J. L.; Bigler, E. D.; Brown, G.; Taylor, J. D. & Rose, S. E. The
        average pathlength map: A diffusion MRI tractography-derived index
        for studying brain pathology. NeuroImage, 2011, 55, 133-141

      * If using -dixel option with TDI contrast only: Smith, R.E., Tournier,
        J-D., Calamante, F., Connelly, A. A novel paradigm for automated
        segmentation of very large whole-brain probabilistic tractography
        data sets. In proc. ISMRM, 2011, 19, 673

      * If using -dixel option with any other contrast: Pannek, K., Raffelt,
        D., Salvado, O., Rose, S. Incorporating directional information in
        diffusion tractography derived maps: angular track imaging (ATI).
        In Proc. ISMRM, 2012, 20, 1912

      * If using -tod option: Dhollander, T., Emsell, L., Van Hecke, W., Maes,
        F., Sunaert, S., Suetens, P. Track Orientation Density Imaging (TODI)
        and Track Orientation Distribution (TOD) based tractography.
        NeuroImage, 2014, 94, 312-336

      * If using other contrasts / statistics: Calamante, F.; Tournier, J.-D.;
        Smith, R. E. & Connelly, A. A generalised framework for
        super-resolution track-weighted imaging. NeuroImage, 2012, 59,
        2494-2503

      * If using -precise mapping option: Smith, R. E.; Tournier, J.-D.;
        Calamante, F. & Connelly, A. SIFT: Spherical-deconvolution informed
        filtering of tractograms. NeuroImage, 2013, 67, 298-312 (Appendix 3)



    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> tdi = mrt.ComputeTDI()
    >>> tdi.inputs.in_file = 'dti.mif'
    >>> tdi.cmdline                               # doctest: +ELLIPSIS
    'tckmap dti.mif tdi.mif'
    >>> tdi.run()                                 # doctest: +SKIP
    """

    _cmd = "tckmap"
    input_spec = ComputeTDIInputSpec
    output_spec = ComputeTDIOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class TCK2VTKInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-2, desc="input tractography"
    )
    out_file = File(
        "tracks.vtk", argstr="%s", usedefault=True, position=-1, desc="output VTK file"
    )
    reference = File(
        exists=True,
        argstr="-image %s",
        desc="if specified, the properties of"
        " this image will be used to convert track point positions from real "
        "(scanner) coordinates into image coordinates (in mm).",
    )
    voxel = File(
        exists=True,
        argstr="-image %s",
        desc="if specified, the properties of"
        " this image will be used to convert track point positions from real "
        "(scanner) coordinates into image coordinates.",
    )

    nthreads = traits.Int(
        argstr="-nthreads %d",
        desc="number of threads. if zero, the number" " of available cpus will be used",
        nohash=True,
    )


class TCK2VTKOutputSpec(TraitedSpec):
    out_file = File(desc="output VTK file")


class TCK2VTK(MRTrix3Base):
    """
    Convert a track file to a vtk format, cave: coordinates are in XYZ
    coordinates not reference

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> vtk = mrt.TCK2VTK()
    >>> vtk.inputs.in_file = 'tracks.tck'
    >>> vtk.inputs.reference = 'b0.nii'
    >>> vtk.cmdline                               # doctest: +ELLIPSIS
    'tck2vtk -image b0.nii tracks.tck tracks.vtk'
    >>> vtk.run()                                 # doctest: +SKIP
    """

    _cmd = "tck2vtk"
    input_spec = TCK2VTKInputSpec
    output_spec = TCK2VTKOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class DWIExtractInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-2, desc="input image"
    )
    out_file = File(argstr="%s", mandatory=True, position=-1, desc="output image")
    bzero = traits.Bool(argstr="-bzero", desc="extract b=0 volumes")
    nobzero = traits.Bool(argstr="-no_bzero", desc="extract non b=0 volumes")
    singleshell = traits.Bool(
        argstr="-singleshell", desc="extract volumes with a specific shell"
    )
    shell = traits.List(
        traits.Float,
        sep=",",
        argstr="-shell %s",
        desc="specify one or more gradient shells",
    )


class DWIExtractOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="output image")


class DWIExtract(MRTrix3Base):
    """
    Extract diffusion-weighted volumes, b=0 volumes, or certain shells from a
    DWI dataset

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> dwiextract = mrt.DWIExtract()
    >>> dwiextract.inputs.in_file = 'dwi.mif'
    >>> dwiextract.inputs.bzero = True
    >>> dwiextract.inputs.out_file = 'b0vols.mif'
    >>> dwiextract.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> dwiextract.cmdline                             # doctest: +ELLIPSIS
    'dwiextract -bzero -fslgrad bvecs bvals dwi.mif b0vols.mif'
    >>> dwiextract.run()                               # doctest: +SKIP
    """

    _cmd = "dwiextract"
    input_spec = DWIExtractInputSpec
    output_spec = DWIExtractOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class MRConvertInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-2, desc="input image"
    )
    out_file = File(
        "dwi.mif",
        argstr="%s",
        mandatory=True,
        position=-1,
        usedefault=True,
        desc="output image",
    )
    coord = traits.List(
        traits.Float,
        sep=" ",
        argstr="-coord %s",
        desc="extract data at the specified coordinates",
    )
    vox = traits.List(
        traits.Float, sep=",", argstr="-vox %s", desc="change the voxel dimensions"
    )
    axes = traits.List(
        traits.Int,
        sep=",",
        argstr="-axes %s",
        desc="specify the axes that will be used",
    )
    scaling = traits.List(
        traits.Float,
        sep=",",
        argstr="-scaling %s",
        desc="specify the data scaling parameter",
    )


class MRConvertOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="output image")


class MRConvert(MRTrix3Base):
    """
    Perform conversion between different file types and optionally extract a
    subset of the input image

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> mrconvert = mrt.MRConvert()
    >>> mrconvert.inputs.in_file = 'dwi.nii.gz'
    >>> mrconvert.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> mrconvert.cmdline                             # doctest: +ELLIPSIS
    'mrconvert -fslgrad bvecs bvals dwi.nii.gz dwi.mif'
    >>> mrconvert.run()                               # doctest: +SKIP
    """

    _cmd = "mrconvert"
    input_spec = MRConvertInputSpec
    output_spec = MRConvertOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class MRMathInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-3, desc="input image"
    )
    out_file = File(argstr="%s", mandatory=True, position=-1, desc="output image")
    operation = traits.Enum(
        "mean",
        "median",
        "sum",
        "product",
        "rms",
        "norm",
        "var",
        "std",
        "min",
        "max",
        "absmax",
        "magmax",
        argstr="%s",
        position=-2,
        mandatory=True,
        desc="operation to computer along a specified axis",
    )
    axis = traits.Int(
        0, argstr="-axis %d", desc="specfied axis to perform the operation along"
    )


class MRMathOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="output image")


class MRMath(MRTrix3Base):
    """
    Compute summary statistic on image intensities
    along a specified axis of a single image

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> mrmath = mrt.MRMath()
    >>> mrmath.inputs.in_file = 'dwi.mif'
    >>> mrmath.inputs.operation = 'mean'
    >>> mrmath.inputs.axis = 3
    >>> mrmath.inputs.out_file = 'dwi_mean.mif'
    >>> mrmath.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> mrmath.cmdline                             # doctest: +ELLIPSIS
    'mrmath -axis 3 -fslgrad bvecs bvals dwi.mif mean dwi_mean.mif'
    >>> mrmath.run()                               # doctest: +SKIP
    """

    _cmd = "mrmath"
    input_spec = MRMathInputSpec
    output_spec = MRMathOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class MRResizeInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True, argstr="%s", position=-2, mandatory=True, desc="input DWI image"
    )
    image_size = traits.Tuple(
        (traits.Int, traits.Int, traits.Int),
        argstr="-size %d,%d,%d",
        mandatory=True,
        desc="Number of voxels in each dimension of output image",
        xor=["voxel_size", "scale_factor"],
    )
    voxel_size = traits.Tuple(
        (traits.Float, traits.Float, traits.Float),
        argstr="-voxel %g,%g,%g",
        mandatory=True,
        desc="Desired voxel size in mm for the output image",
        xor=["image_size", "scale_factor"],
    )
    scale_factor = traits.Tuple(
        (traits.Float, traits.Float, traits.Float),
        argstr="-scale %g,%g,%g",
        mandatory=True,
        desc="Scale factors to rescale the image by in each dimension",
        xor=["image_size", "voxel_size"],
    )
    interpolation = traits.Enum(
        "cubic",
        "nearest",
        "linear",
        "sinc",
        argstr="-interp %s",
        usedefault=True,
        desc="set the interpolation method to use when resizing (choices: "
        "nearest, linear, cubic, sinc. Default: cubic).",
    )
    out_file = File(
        argstr="%s",
        name_template="%s_resized",
        name_source=["in_file"],
        keep_extension=True,
        position=-1,
        desc="the output resized DWI image",
    )


class MRResizeOutputSpec(TraitedSpec):
    out_file = File(desc="the output resized DWI image", exists=True)


class MRResize(MRTrix3Base):
    """
    Resize an image by defining the new image resolution, voxel size or a
    scale factor. If the image is 4D, then only the first 3 dimensions can be
    resized. Also, if the image is down-sampled, the appropriate smoothing is
    automatically applied using Gaussian smoothing.
    For more information, see
    <https://mrtrix.readthedocs.io/en/latest/reference/commands/mrresize.html>

    Example
    -------
    >>> import nipype.interfaces.mrtrix3 as mrt

    Defining the new image resolution:
    >>> image_resize = mrt.MRResize()
    >>> image_resize.inputs.in_file = 'dwi.mif'
    >>> image_resize.inputs.image_size = (256, 256, 144)
    >>> image_resize.cmdline                               # doctest: +ELLIPSIS
    'mrresize -size 256,256,144 -interp cubic dwi.mif dwi_resized.mif'
    >>> image_resize.run()                                 # doctest: +SKIP

    Defining the new image's voxel size:
    >>> voxel_resize = mrt.MRResize()
    >>> voxel_resize.inputs.in_file = 'dwi.mif'
    >>> voxel_resize.inputs.voxel_size = (1, 1, 1)
    >>> voxel_resize.cmdline                               # doctest: +ELLIPSIS
    'mrresize -interp cubic -voxel 1,1,1 dwi.mif dwi_resized.mif'
    >>> voxel_resize.run()                                 # doctest: +SKIP

    Defining the scale factor of each image dimension:
    >>> scale_resize = mrt.MRResize()
    >>> scale_resize.inputs.in_file = 'dwi.mif'
    >>> scale_resize.inputs.scale_factor = (0.5,0.5,0.5)
    >>> scale_resize.cmdline                               # doctest: +ELLIPSIS
    'mrresize -interp cubic -scale 0.5,0.5,0.5 dwi.mif dwi_resized.mif'
    >>> scale_resize.run()                                 # doctest: +SKIP
    """

    _cmd = "mrresize"
    input_spec = MRResizeInputSpec
    output_spec = MRResizeOutputSpec



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..ukftractography import UKFTractography


def test_UKFTractography_inputs():
    input_map = dict(
        Ql=dict(
            argstr="--Ql %f",
        ),
        Qm=dict(
            argstr="--Qm %f",
        ),
        Qw=dict(
            argstr="--Qw %f",
        ),
        Rs=dict(
            argstr="--Rs %f",
        ),
        args=dict(
            argstr="%s",
        ),
        dwiFile=dict(
            argstr="--dwiFile %s",
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        freeWater=dict(
            argstr="--freeWater ",
        ),
        fullTensorModel=dict(
            argstr="--fullTensorModel ",
        ),
        labels=dict(
            argstr="--labels %s",
            sep=",",
        ),
        maskFile=dict(
            argstr="--maskFile %s",
            extensions=None,
        ),
        maxBranchingAngle=dict(
            argstr="--maxBranchingAngle %f",
        ),
        maxHalfFiberLength=dict(
            argstr="--maxHalfFiberLength %f",
        ),
        minBranchingAngle=dict(
            argstr="--minBranchingAngle %f",
        ),
        minFA=dict(
            argstr="--minFA %f",
        ),
        minGA=dict(
            argstr="--minGA %f",
        ),
        numTensor=dict(
            argstr="--numTensor %s",
        ),
        numThreads=dict(
            argstr="--numThreads %d",
        ),
        recordCovariance=dict(
            argstr="--recordCovariance ",
        ),
        recordFA=dict(
            argstr="--recordFA ",
        ),
        recordFreeWater=dict(
            argstr="--recordFreeWater ",
        ),
        recordLength=dict(
            argstr="--recordLength %f",
        ),
        recordNMSE=dict(
            argstr="--recordNMSE ",
        ),
        recordState=dict(
            argstr="--recordState ",
        ),
        recordTensors=dict(
            argstr="--recordTensors ",
        ),
        recordTrace=dict(
            argstr="--recordTrace ",
        ),
        seedFALimit=dict(
            argstr="--seedFALimit %f",
        ),
        seedsFile=dict(
            argstr="--seedsFile %s",
            extensions=None,
        ),
        seedsPerVoxel=dict(
            argstr="--seedsPerVoxel %d",
        ),
        stepLength=dict(
            argstr="--stepLength %f",
        ),
        storeGlyphs=dict(
            argstr="--storeGlyphs ",
        ),
        tracts=dict(
            argstr="--tracts %s",
            hash_files=False,
        ),
        tractsWithSecondTensor=dict(
            argstr="--tractsWithSecondTensor %s",
            hash_files=False,
        ),
        writeAsciiTracts=dict(
            argstr="--writeAsciiTracts ",
        ),
        writeUncompressedTracts=dict(
            argstr="--writeUncompressedTracts ",
        ),
    )
    inputs = UKFTractography.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_UKFTractography_outputs():
    output_map = dict(
        tracts=dict(
            extensions=None,
        ),
        tractsWithSecondTensor=dict(
            extensions=None,
        ),
    )
    outputs = UKFTractography.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Descriptor support for NIPY.

Utilities to support special Python descriptors [1,2], in particular the use of
a useful pattern for properties we call 'one time properties'.  These are
object attributes which are declared as properties, but become regular
attributes once they've been read the first time.  They can thus be evaluated
later in the object's life cycle, but once evaluated they become normal, static
attributes with no function call overhead on access or any other constraints.

References
----------
[1] How-To Guide for Descriptors, Raymond
Hettinger. http://users.rcn.com/python/download/Descriptor.htm

[2] Python data model, http://docs.python.org/reference/datamodel.html
"""


class OneTimeProperty(object):
    """A descriptor to make special properties that become normal attributes.
    """

    def __init__(self, func):
        """Create a OneTimeProperty instance.

         Parameters
         ----------
           func : method

             The method that will be called the first time to compute a value.
             Afterwards, the method's name will be a standard attribute holding
             the value of this computation.
        """
        self.getter = func
        self.name = func.__name__

    def __get__(self, obj, type=None):
        """ Called on attribute access on the class or instance.  """
        if obj is None:
            # Being called on the class, return the original function.
            # This way, introspection works on the class.
            return self.getter

        val = self.getter(obj)
        # print "** setattr_on_read - loading '%s'" % self.name  # dbg
        setattr(obj, self.name, val)
        return val


def setattr_on_read(func):
    # XXX - beetter names for this?
    # - cor_property (copy on read property)
    # - sor_property (set on read property)
    # - prop2attr_on_read
    # ... ?
    """Decorator to create OneTimeProperty attributes.

    Parameters
    ----------
      func : method
        The method that will be called the first time to compute a value.
        Afterwards, the method's name will be a standard attribute holding the
        value of this computation.

    Examples
    --------
    >>> class MagicProp(object):
    ...     @setattr_on_read
    ...     def a(self):
    ...         return 99
    ...
    >>> x = MagicProp()
    >>> 'a' in x.__dict__
    False
    >>> x.a
    99
    >>> 'a' in x.__dict__
    True
    """
    return OneTimeProperty(func)



# -*- coding: utf-8 -*-
import os
import argparse
import inspect
import sys

from ..interfaces.base import Interface, InputMultiPath, traits
from ..interfaces.base.support import get_trait_desc
from .misc import str2bool


def listClasses(module=None):
    if module:
        __import__(module)
        pkg = sys.modules[module]
        print("Available Interfaces:")
        for k, v in sorted(list(pkg.__dict__.items())):
            if inspect.isclass(v) and issubclass(v, Interface):
                print("\t%s" % k)


def add_options(parser=None, module=None, function=None):
    interface = None
    if parser and module and function:
        __import__(module)
        interface = getattr(sys.modules[module], function)()

        inputs = interface.input_spec()
        for name, spec in sorted(interface.inputs.traits(transient=None).items()):
            desc = "\n".join(get_trait_desc(inputs, name, spec))[len(name) + 2 :]
            args = {}

            if spec.is_trait_type(traits.Bool):
                args["action"] = "store_true"

            if hasattr(spec, "mandatory") and spec.mandatory:
                if spec.is_trait_type(InputMultiPath):
                    args["nargs"] = "+"
                parser.add_argument(name, help=desc, **args)
            else:
                if spec.is_trait_type(InputMultiPath):
                    args["nargs"] = "*"
                parser.add_argument("--%s" % name, dest=name, help=desc, **args)
    return parser, interface


def run_instance(interface, options):
    print("setting function inputs")

    for input_name, _ in list(interface.inputs.items()):
        if getattr(options, input_name) is not None:
            value = getattr(options, input_name)
            try:
                setattr(interface.inputs, input_name, value)
            except ValueError as e:
                print("Error when setting the value of %s: '%s'" % (input_name, str(e)))

    print(interface.inputs)
    res = interface.run()
    print(res.outputs)


def main(argv):

    if len(argv) == 2 and not argv[1].startswith("-"):
        listClasses(argv[1])
        sys.exit(0)

    parser = argparse.ArgumentParser(
        description="Nipype interface runner", prog=argv[0]
    )
    parser.add_argument("module", type=str, help="Module name")
    parser.add_argument("interface", type=str, help="Interface name")
    parsed = parser.parse_args(args=argv[1:3])

    _, prog = os.path.split(argv[0])
    interface_parser = argparse.ArgumentParser(
        description="Run %s" % parsed.interface, prog=" ".join([prog] + argv[1:3])
    )
    interface_parser, interface = add_options(
        interface_parser, parsed.module, parsed.interface
    )
    args = interface_parser.parse_args(args=argv[3:])
    run_instance(interface, args)
